from __future__ import division
from past.utils import old_div
import numpy as np


def angular_distance_fast(ra1, dec1, ra2, dec2):
    """
    Compute angular distance using the Haversine formula. Use this one when you know you will never ask for points at
    their antipodes. If this is not the case, use the angular_distance function which is slower, but works also for
    antipodes.

    :param lon1:
    :param lat1:
    :param lon2:
    :param lat2:
    :return:
    """

    lon1 = np.deg2rad(ra1)
    lat1 = np.deg2rad(dec1)
    lon2 = np.deg2rad(ra2)
    lat2 = np.deg2rad(dec2)
    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = np.sin(old_div(dlat,2.0))**2 + np.cos(lat1) * np.cos(lat2) * np.sin(old_div(dlon,2.0))**2
    c = 2 * np.arcsin(np.sqrt(a))
    return np.rad2deg(c)


def angular_distance(ra1, dec1, ra2, dec2):
    """
    Returns the angular distance between two points, two sets of points, or a set of points and one point.

    :param ra1: array or float, longitude of first point(s)
    :param dec1: array or float, latitude of first point(s)
    :param ra2: array or float, longitude of second point(s)
    :param dec2: array or float, latitude of second point(s)
    :return: angular distance(s) in degrees
    """

    # Vincenty formula, slower than the Haversine formula in some cases, but stable also at antipodes

    lon1 = np.deg2rad(ra1)
    lat1 = np.deg2rad(dec1)
    lon2 = np.deg2rad(ra2)
    lat2 = np.deg2rad(dec2)

    sdlon = np.sin(lon2 - lon1)
    cdlon = np.cos(lon2 - lon1)
    slat1 = np.sin(lat1)
    slat2 = np.sin(lat2)
    clat1 = np.cos(lat1)
    clat2 = np.cos(lat2)

    num1 = clat2 * sdlon
    num2 = clat1 * slat2 - slat1 * clat2 * cdlon
    denominator = slat1 * slat2 + clat1 * clat2 * cdlon

    return np.rad2deg(np.arctan2(np.sqrt(num1 ** 2 + num2 ** 2), denominator))


def spherical_angle( ra0, dec0, ra1, dec1, ra2, dec2 ):
    """
    Returns the spherical angle distance between two sets of great circles defined by (ra0, dec0), (ra1, dec1) and (ra0, dec0), (ra2, dec2)

    :param ra0: array or float, longitude of intersection point(s)
    :param dec0: array or float, latitude of intersection point(s)
    :param ra1: array or float, longitude of first point(s)
    :param dec1: array or float, latitude of first point(s)
    :param ra2: array or float, longitude of second point(s)
    :param dec2: array or float, latitude of second point(s)
    :return: spherical angle in degrees
    """
    
    a = np.deg2rad( angular_distance(ra0, dec0, ra1, dec1))
    b = np.deg2rad( angular_distance(ra0, dec0, ra2, dec2))
    c = np.deg2rad( angular_distance(ra2, dec2, ra1, dec1))
    
    #use the spherical law of cosines: https://en.wikipedia.org/wiki/Spherical_law_of_cosines#Rearrangements
    
    numerator = np.atleast_1d( np.cos(c) - np.cos(a) * np.cos(b) )
    denominator = np.atleast_1d( np.sin(a)*np.sin(b) )
    
    return np.where( denominator == 0 , np.zeros( len(denominator)), np.rad2deg( np.arccos( old_div(numerator,denominator))) )



from .functions import (Blackbody, DiracDelta, Exponential_cutoff,
                        Log_parabola, Sin, StepFunction, StepFunctionUpper,
                        has_gsl, has_naima)

if has_naima:
    from .functions import Synchrotron

if has_gsl:

    from .functions import Cutoff_powerlaw_flux

from .absorption import PhAbs, TbAbs, WAbs, has_ebltable
from .apec import has_atomdb

if has_ebltable:
    from .absorption import EBLattenuation

from .polynomials import (Constant, Cubic, Line, Quadratic, Quartic,
                          get_polynomial)
from .powerlaws import (Band, Band_Calderone, Band_grbm, Broken_powerlaw,
                        Cutoff_powerlaw, Cutoff_powerlaw_Ep, Inverse_cutoff_powerlaw, Powerlaw,
                        Powerlaw_Eflux, Powerlaw_flux, SmoothlyBrokenPowerLaw,
                        Super_cutoff_powerlaw)

if has_atomdb:

    from .apec import APEC, VAPEC



__all__ = ["Band", "Band_Calderone", "Band_grbm", "Broken_powerlaw",
           "Cutoff_powerlaw", "Cutoff_powerlaw_Ep", "Inverse_cutoff_powerlaw", "Powerlaw",
           "Powerlaw_Eflux", "Powerlaw_flux", "SmoothlyBrokenPowerLaw",
           "Super_cutoff_powerlaw",
           "Constant", "Cubic", "DiracDelta", "Exponential_cutoff", "Line",
           "Quadratic", "Sin", "StepFunction", "StepFunctionUpper",
           "PhAbs", "TbAbs", "WAbs",
           "Log_parabola",
           "Blackbody","Quartic", "get_polynomial"
           ]

if has_atomdb:
    __all__.extend(["APEC", "VAPEC"])


if has_gsl:

    __all__.extend(["Cutoff_powerlaw_flux"])


if has_naima:

    __all__.extend(["Synchrotron"])


if has_ebltable:

    __all__.extend(["EBLattenuation"])



from __future__ import absolute_import, division

from builtins import range, str

from past.utils import old_div

__author__ = "giacomov"

__doc__ = """"""

import collections
import copy
import warnings
from typing import Optional, Union, List, Dict, Any

import astropy.units as u
import numpy as np
import scipy.stats

from astromodels.core.parameter_transformation import ParameterTransformation
from astromodels.utils.logging import setup_logger

from .thread_safe_unit_format import ThreadSafe
from .tree import Node

log = setup_logger(__name__)


def _behaves_like_a_number(obj) -> True:
    """

    :param obj:
    :return: True if obj supports addition, subtraction, multiplication and division. False otherwise.
    """

    try:

        obj + 1
        obj * 2
        old_div(obj, 2)
        obj - 1

    except TypeError:

        return False

    else:

        return True


# Exception for when a parameter is out of its bounds
class SettingOutOfBounds(RuntimeError):
    pass


# Exception for when an object should be callable but it isn't
class NotCallableOrErrorInCall(RuntimeError):
    pass


class WarningUnitsAreSlow(Warning):
    pass


class IndependentVariableCannotBeLinked(RuntimeError):
    pass


class CannotUnderstandUnit(RuntimeError):
    pass


class CannotConvertValueToNewUnits(RuntimeError):
    pass


class ParameterMustHaveBounds(RuntimeError):
    pass


def accept_quantity(input_type=float, allow_none=False):
    """
    A class-method decorator which allow a given method (typically the set_value method) to receive both a
    astropy.Quantity or a simple float, but to be coded like it's always receiving a pure float in the right units.
    This is to give a way to avoid the huge bottleneck that are astropy.units

    :param input_type: the expected type for the input (float, int)
    :param allow_none : whether to allow or not the passage of None as argument (default: False)
    :return: a decorator for the particular type
    """
    def accept_quantity_wrapper(method):
        def handle_quantity(instance, value, *args, **kwargs):

            # For speed reasons, first run the case where the input is not a quantity, and fall back to the handling
            # of quantities if that fails. The parts that fails if the input is a Quantity is the conversion
            # input_type(value). This could have been handled more elegantly with a "finally" clause, but that would
            # have a 40 percent speed impact...

            try:

                new_value = input_type(value)

                return method(instance, new_value, *args, **kwargs)

            except TypeError:

                # Slow for slow, check that we actually have a quantity or None (if allowed)

                if isinstance(value, u.Quantity):

                    new_value = value.to(instance.unit).value

                    return method(instance, new_value, *args, **kwargs)

                elif value is None:

                    if allow_none:

                        return method(instance, None, *args, **kwargs)

                    else:  # pragma: no cover

                        log.exception("You cannot pass None as argument for "
                                      "method %s of %s" %
                                      (method.__name__, instance.name))

                        raise TypeError()

                else:  # pragma: no cover

                    log.exception(
                        "You need to pass either a %s or a astropy.Quantity "
                        "to method %s of %s" %
                        (input_type.__name__, method.__name__, instance.name))

                    raise TypeError()

        return handle_quantity

    return accept_quantity_wrapper


class ParameterBase(Node):
    """

    :param name: name for parameter
    :param value: initial value
    :param min_value: minimum
    :param max_value: maximum
    :param desc: description
    :param unit: units (string or astropy.Unit)
    :param transformation: a class which implements a .forward and a .backward method to transform forth and back from
        face value (the value exposed to the user) to the internal value (the value exposed to the fitting engine)
    """
    def __init__(
        self,
        name: str,
        value: float,
        min_value: Optional[float]=None,
        max_value: Optional[float]=None,
        desc: Optional[str]=None,
        unit: u.Unit=u.dimensionless_unscaled,
        transformation: Optional[ParameterTransformation]=None,
    ):

        # Make this a node

        Node.__init__(self, name)

        # Make a static name which will never change (not even after a _change_name call)
        self._static_name: str = str(name)

        # Callbacks are executed any time the value for the parameter changes (i.e., its value changes)

        # We start from a empty list of callbacks.
        self._callbacks = []

        # Assign to members

        # Store the units as an astropy.units.Unit instance

        self._unit = self._safe_assign_unit(unit)

        # A ParameterBase instance cannot have auxiliary variables
        self._aux_variable: Dict[str, Any] = {}

        # Set min and max to None first so that the .value setter will work,
        # we will override them later if needed
        self._external_min_value: Optional[float] = None
        self._external_max_value: Optional[float] = None

        # Store the transformation. This allows to disentangle the value of the parameter which the user interact
        # width with the value of the parameter the fitting engine (or the Bayesian sampler) interact with
        if transformation is not None:
            if not isinstance(transformation, ParameterTransformation):

                log.error("transformation is not of ParameterTransform ")

                raise AssertionError()
                
        self._transformation: Optional[ParameterTransformation] = transformation

        # Let's store the init value

        # NOTE: this will be updated immediately by the _set_value method of the "value" property
        self._internal_value: Optional[float] = None
        # If the value is a Quantity, deal with that

        if isinstance(value, u.Quantity):

            # If the user did not specify an ad-hoc unit, use the unit
            # of the Quantity

            if self._unit == u.dimensionless_unscaled:

                self._unit = value.unit

            # Convert the value to the provided unit (if necessary)

            self.value: float = value.to(self._unit).value

        else:

            self.value = value

        # Set minimum if provided, otherwise use default
        # (use the property so the checks that are there are performed also on construction)

        self.min_value: Optional[float] = min_value

        # Set maximum if provided, otherwise use default

        # this will be overwritten immediately in the next line
        self.max_value: Optional[float] = max_value

        # Store description

        self._desc: Optional[float] = desc

        # Make the description the documentation as well

        self.__doc__ = desc

        # Now perform a very lazy check that we can perform math operations on value, minimum and maximum
        # (i.e., they are numbers)

        if not _behaves_like_a_number(self.value):

            log.exception("The provided initial value is not a number")

            raise TypeError()

    def _repr__base(self, rich_output):  # pragma: no cover

        raise NotImplementedError(
            "You need to implement this for the actual Parameter class"
        )

    # Define the property 'description' and make it read-only

    @property
    def static_name(self) -> str:
        """
        Returns a name which will never change, even if the name of the parameter does (for example in composite
        functions)

        :return : the static name
        :type : str
        """

        return self._static_name

    @property
    def description(self) -> Optional[str]:
        """
        Return a description of this parameter

        :return: a string cointaining a description of the meaning of this parameter
        """
        return self._desc

    @staticmethod
    def _safe_assign_unit(input_unit) -> u.Unit:

        # We first try to use our own, thread-safe format, if we fail then we try the astropy one

        try:

            new_unit = u.Unit(input_unit, format="threadsafe")

        except ValueError:

            # Try with the default format of astropy
            new_unit = u.Unit(input_unit)

        return new_unit

    # Define the property 'unit'
    def _set_unit(self, input_unit) -> None:

        # This will fail if the input is not valid

        new_unit = self._safe_assign_unit(input_unit)

        # Now transform the current _value in the new unit, unless the current unit is dimensionless, in which
        # case there is no transformation to make

        # (self._unit is the OLD unit here)

        if self._unit != u.dimensionless_unscaled:

            # This will fail if the new unit is not compatible with the old one

            try:

                self.value = self.as_quantity.to(new_unit).value

            except u.UnitConversionError:

                if new_unit == u.dimensionless_unscaled:

                    new_unit_name = "(dimensionless)"

                else:

                    new_unit_name = new_unit

                log.exception(
                    "Cannot convert the value %s from %s to the "
                    "new units %s" % (self.value, self._unit, new_unit_name)
                )

                raise CannotConvertValueToNewUnits()

        else:

            # This is possibly the first time the unit is set
            pass

        # Finally store the new unit

        self._unit = new_unit

    def _get_unit(self):

        return self._unit

    unit = property(
        _get_unit, _set_unit, doc="""Gets or sets the unit for the parameter"""
    )

    @property
    def as_quantity(self) -> u.Quantity:
        """
        Return the current value with its units (as an astropy.Quantity instance)

        :return: an instance of astropy.Quantity)
        """
        return self.value * self._unit

    def in_unit_of(self, unit, as_quantity=False) -> u.Quantity:
        """
        Return the current value transformed to the new units

        :param unit: either an astropy.Unit instance, or a string which can be converted to an astropy.Unit
            instance, like "1 / (erg cm**2 s)"
        :param as_quantity: if True, the method return an astropy.Quantity, if False just a floating point number.
            Default is False
        :return: either a floating point or a astropy.Quantity depending on the value of "as_quantity"
        """

        new_unit = u.Unit(unit)

        new_quantity = self.as_quantity.to(new_unit)

        if as_quantity:

            return new_quantity

        else:

            return new_quantity.value

    def has_auxiliary_variable(self) -> bool:

        if self._aux_variable:

            return True

        else:

            return False

    def has_transformation(self) -> bool:

        if self._transformation is None:

            return False

        else:

            return True

    @property
    def transformation(self) -> Optional[ParameterTransformation]:

        return self._transformation

    def internal_to_external_delta(self, internal_value, internal_delta):
        """
        Transform an interval from the internal to the external reference (through the transformation). It is useful
        if you have for example a confidence interval in internal reference and you want to transform it to the
        external reference

        :param interval_value: value in internal reference
        :param internal_delta: delta in internal reference
        :return: value and delta in external reference
        """

        external_value = self.transformation.backward(internal_value)
        bound_internal = internal_value + internal_delta
        bound_external = self.transformation.backward(bound_internal)
        external_delta = bound_external - external_value

        return external_value, external_delta

    # Define the property "value" with a control that the parameter cannot be set
    # outside of its bounds

    def _get_value(self) -> float:
        """Return current parameter value"""

        # This is going to be true (possibly) only for derived classes. It is here to make the code cleaner
        # and also to avoid infinite recursion

        if self._aux_variable:

            return self._aux_variable["law"](self._aux_variable["variable"].value)

        if self._transformation is None:

            return self._internal_value

        else:

            # A transformation is set. Transform back from internal value to true value
            #
            # print("Interval value is %s" % self._internal_value)
            # print("Returning %s" % self._transformation.backward(self._internal_value))

            return self._transformation.backward(self._internal_value)

    # NOTE: this function should only be used by the user. Fitting engines should only deal with
    # _get_internal_value and _set_internal_value
    @accept_quantity(
        float, allow_none=False
    )  # This means that the method will always receive a float
    def _set_value(self, new_value):
        """Sets the current value of the parameter, ensuring that it is within the allowed range."""

        if self.min_value is not None and new_value < self.min_value:

            raise SettingOutOfBounds(
                "Trying to set parameter {0} = {1}, which is less than the minimum allowed {2}".format(
                    self.name, new_value, self.min_value
                )
            )

        if self.max_value is not None and new_value > self.max_value:
            raise SettingOutOfBounds(
                "Trying to set parameter {0} = {1}, which is more than the maximum allowed {2}".format(
                    self.name, new_value, self.max_value
                )
            )

        # Issue a warning if there is an auxiliary variable, as the setting does not have any effect
        if self.has_auxiliary_variable():

            with warnings.catch_warnings():

                warnings.simplefilter("always", RuntimeWarning)

                log.warning(
                    "You are trying to assign to a parameter which is either linked or "
                    "has auxiliary variables. The assignment has no effect."
                )

        # Save the value as a pure floating point to avoid the overhead of the astropy.units machinery when
        # not needed

        if self._transformation is None:

            new_internal_value = new_value

        else:

            new_internal_value = self._transformation.forward(new_value)

        # If the parameter has changed, update its value and call the callbacks if needed

        if new_internal_value != self._internal_value:

            # Update
            self._internal_value = new_internal_value

            # Call the callbacks (if any)
            for callback in self._callbacks:

                try:

                    callback(self)

                except:

                    log.exception(
                        "Could not call callback for parameter %s" % self.name
                    )

                    raise NotCallableOrErrorInCall()

    value = property(
        _get_value,
        _set_value,
        doc="Get and sets the current value for the parameter, with or without units",
    )

    def _get_internal_value(self):
        """
        This is supposed to be only used by fitting engines at the beginning to get the starting value for free
        parameters. From then on, only the _set_internal_value should be used

        :return: the internal value
        """

        # NOTE: we don't need here to deal with auxiliary variables because if one is defined, the parameter is not
        # free thus it will not be touched by the fitting engine
        assert len(self._aux_variable) == 0, (
            "You cannot get the internal value of a parameter which has an auxiliary "
            "variable"
        )

        return self._internal_value

    def _set_internal_value(self, new_internal_value):
        """
        This is supposed to be only used by fitting engines

        :param new_internal_value: new value in internal representation
        :return: none
        """

        if new_internal_value != self._internal_value:

            self._internal_value = new_internal_value

            # Call callbacks if any

            for callback in self._callbacks:

                callback(self)

    # Define the property "min_value"
    # NOTE: the min value is always in external representation

    def _get_min_value(self):
        """Return current minimum allowed value"""

        return self._external_min_value

    @accept_quantity(float, allow_none=True)
    def _set_min_value(self, min_value):

        """Sets current minimum allowed value"""

        # Check that the min value can be transformed if a transformation is present
        if self._transformation is not None:

            if min_value is not None:

                if self._transformation.is_positive:

                    assert min_value > 0.0, (
                        "The transformation %s is postive definite and the min_value was set to a negative number for %s "
                        % (type(self._transformation), self.path)
                    )

                try:

                    _ = self._transformation.forward(min_value)

                except FloatingPointError:

                    log.exception(
                        "The provided minimum %s cannot be transformed with the transformation %s which "
                        "is defined for the parameter %s"
                        % (min_value, type(self._transformation), self.path)
                    )

                    raise ValueError()
            else:

                if self._transformation.is_positive:

                    # set it by default to for the user
                    min_value = 1e-99

                    log.warning(
                        "We have set the min_value of %s to 1e-99 because there was a postive transform"
                        % self.path
                    )

        # Store the minimum as a pure float

        self._external_min_value = min_value

        # Check that the current value of the parameter is still within the boundaries. If not, issue a warning

        if (
            self._external_min_value is not None
            and self.value < self._external_min_value
        ):

            log.warning(
                "The current value of the parameter %s (%s) "
                "was below the new minimum %s."
                % (self.name, self.value, self._external_min_value)
            )

            self.value = self._external_min_value

    min_value = property(
        _get_min_value,
        _set_min_value,
        doc="Gets or sets the minimum allowed value for the parameter",
    )

    def remove_minimum(self):
        """
        Remove the minimum from this parameter (i.e., it becomes boundless in the negative direction)
        """
        self._external_min_value = None

    def _set_internal_min_value(self):

        log.exception(
            "You should never attempt to change the internal representation of the minimum"
        )

        raise NotCallableOrErrorInCall()

    def _get_internal_min_value(self):
        """
        This is supposed to be only used by fitting engines to get the minimum value in internal representation.
        It is supposed to be called only once before doing the minimization/sampling, to set the range of the parameter

        :return: minimum value in internal representation (or None if there is no minimum)
        """

        if self.min_value is None:

            # No minimum set

            return None

        else:

            # There is a minimum. If there is a transformation, use it, otherwise just return the minimum

            if self._transformation is None:

                return self._external_min_value

            else:

                return self._transformation.forward(self._external_min_value)

    # Define the property "max_value"

    def _get_max_value(self):

        """Return current maximum allowed value"""

        return self._external_max_value

    @accept_quantity(float, allow_none=True)
    def _set_max_value(self, max_value):
        """Sets current maximum allowed value"""

        self._external_max_value = max_value

        # Check that the current value of the parameter is still within the boundaries. If not, issue a warning

        if (
            self._external_max_value is not None
            and self.value > self._external_max_value
        ):

            log.warning(
                "The current value of the parameter %s (%s) "
                "was above the new maximum %s."
                % (self.name, self.value, self._external_max_value)
            )
            self.value = self._external_max_value

    max_value = property(
        _get_max_value,
        _set_max_value,
        doc="Gets or sets the maximum allowed value for the parameter",
    )

    def remove_maximum(self):
        """
        Remove the maximum from this parameter (i.e., it becomes boundless in the positive direction)
        """
        self._external_max_value = None

    def _set_internal_max_value(self):

        log.exception(
            "You should never attempt to change the internal representation of the minimum"
        )

        raise NotCallableOrErrorInCall()

    def _get_internal_max_value(self):
        """
        This is supposed to be only used by fitting engines to get the maximum value in internal representation.
        It is supposed to be called only once before doing the minimization/sampling, to set the range of the parameter

        :return: maximum value in internal representation (or None if there is no minimum)
        """

        if self.max_value is None:

            # No minimum set

            return None

        else:

            # There is a minimum. If there is a transformation, use it, otherwise just return the minimum

            if self._transformation is None:

                return self._external_max_value

            else:

                return self._transformation.forward(self._external_max_value)

    def _set_bounds(self, bounds):
        """Sets the boundaries for this parameter to min_value and max_value"""

        # Use the properties so that the checks and the handling of units are made automatically

        min_value, max_value = bounds

        # Remove old boundaries to avoid problems with the new one, if the current value was within the old boundaries
        # but is not within the new ones (it will then be adjusted automatically later)
        self.min_value = None
        self.max_value = None

        self.min_value = min_value

        self.max_value = max_value

    def _get_bounds(self):
        """Returns the current boundaries for the parameter"""

        return self.min_value, self.max_value

    bounds = property(
        _get_bounds,
        _set_bounds,
        doc="Gets or sets the boundaries (minimum and maximum) for this " "parameter",
    )

    def add_callback(self, callback):
        """Add a callback to the list of functions which are called immediately after the value of the parameter
        is changed. The callback must be a function accepting the current parameter as input. The return value of the
        callback is ignored. More than one callback can be specified. In that case, the callbacks will be called in the
        same order they have been entered."""

        self._callbacks.append(callback)

    def get_callbacks(self):
        """
        Returns the list of callbacks currently defined

        :return:

        """
        return self._callbacks

    def empty_callbacks(self) -> None:
        """Remove all callbacks for this parameter"""
        self._callbacks = []

    def duplicate(self) -> "Parameter":
        """
        Returns an exact copy of the current parameter
        """

        # Deep copy everything to make sure that there are no ties between the new instance and the old one

        new_parameter = copy.deepcopy(self)

        return new_parameter

    def to_dict(self, minimal=False) -> Dict[str, Any]:

        """Returns the representation for serialization"""

        data = collections.OrderedDict()

        if minimal:

            # In the minimal representation we just output the value

            data["value"] = self._to_python_type(self.value)

        else:

            # In the complete representation we output everything is needed to re-build the object

            data["value"] = self._to_python_type(self.value)
            data["desc"] = str(self.description)
            data["min_value"] = self._to_python_type(self.min_value)
            data["max_value"] = self._to_python_type(self.max_value)
            # We use our own thread-safe format for the unit
            data["unit"] = self.unit.to_string(format="threadsafe")

        return data

    @staticmethod
    def _to_python_type(variable):
        """
        Returns the value in the variable handling also np.array of one element

        :param variable: input variable
        :return: the value of the variable having a python type (int, float, ...)
        """

        # Assume variable is a np.array, fall back to the case where variable is already a primitive type

        try:

            return variable.item()

        except AttributeError:

            return variable


class Parameter(ParameterBase):
    """

    Implements a numerical parameter. Optionally the parameter can vary according to an auxiliary law (see below).

    :param name: Name for the parameter
    :param value: Initial value
    :param min_value: minimum allowed value for the parameter (default: None)
    :param max_value: maximum allowed value for the parameter (default: None)
    :param delta: initial step used by some fitting engines (default: 0.1 * value)
    :param desc: description of parameter (default: '')
    :param free: whether the parameter is free or not (default: True)
    :param unit: the parameter units (default: dimensionless)
    :param prior: the parameter's prior (default: None)
    :param is_normalization: True or False, wether the parameter is a normalization or not (default: False)
    :param transformation: a transformation to be used between external value (the value the user interacts with) and
        the value the fitting/sampling engine interacts with (internal value). It is an instance of a class implementing a
        forward(external_value) and a backward(internal_value) method returning respectively the transformation of the
        external value in the internal value and viceversa. This is useful because for example the logarithm of a parameter
        with a large range of possible values (say from 1e-12 to 1e20) is handled much better by fitting engines than the
        raw value. The log transformation indeed makes the gradient much easier to compute.
    """

    def __init__(
        self,
        name: Optional[str]=None,
        value: Optional[float] = None,
        min_value: Optional[float]=None,
        max_value: Optional[float]=None,
        delta: Optional[float]=None,
        desc: Optional[str]=None,
        free: bool=True,
        unit="",
        prior=None,
        is_normalization: bool=False,
        transformation: Optional[ParameterTransformation]=None,
    ):

        # This extends ParameterBase by adding the possibility for free/fix, and a delta for fitting purposes, as
        # well as a prior

        super(Parameter, self).__init__(
            name,
            value,
            min_value=min_value,
            max_value=max_value,
            desc=desc,
            unit=unit,
            transformation=transformation,
        )

        self._free: bool = bool(free)

        # Set delta if provided, otherwise use default

        if delta is not None:

            self._delta: float = delta

        else:

            # Default is 10% of the value, unless the value is zero, in which case the delta is 0.1

            if self.value == 0:

                self._delta = 0.1

            else:

                self._delta = abs(0.1 * self.value)

        # pre-defined prior is no prior
        self._prior = None

        # override the default if a prior has been given

        if prior is not None:

            # Use the property on purpose, so all checks and setups are applied

            self.prior = prior

        # Now perform a very lazy check that we can perform math operations on the delta

        if not _behaves_like_a_number(self._delta):
            log.exception("The provided delta is not a number")
            raise TypeError()

        # Create a backup copy of the status of the parameter (useful when an auxiliary variable
        # is removed)
        self._old_free = self._free

        self._is_normalization = bool(is_normalization)

    @property
    def is_normalization(self) -> bool:

        return self._is_normalization

    # Define the property "delta"

    def _get_delta(self):
        return self._delta

    @accept_quantity(float, allow_none=False)
    def _set_delta(self, delta):
        """Sets the current delta for the parameter. The delta is used as initial step by some fitting engines"""

        self._delta = delta

    delta = property(
        _get_delta, _set_delta, doc="""Gets or sets the delta for the parameter"""
    )

    def _get_internal_delta(self):
        """
        This is only supposed to be used by fitting/sampling engine, to get the initial step in internal representation

        :return: initial delta in internal representation
        """

        if self._transformation is None:

            return self._delta

        else:

            delta_int = None

            for i in range(2):

                # Try using the low bound

                low_bound_ext = self.value - self.delta

                # Make sure we are within the margins

                if low_bound_ext > self.min_value:

                    # Ok, let's use that for the delta
                    low_bound_int = self._transformation.forward(low_bound_ext)

                    delta_int = abs(low_bound_int - self._get_internal_value())

                    break

                else:

                    # Nope, try with the hi bound

                    hi_bound_ext = self.value + self._delta

                    if hi_bound_ext < self.max_value:

                        # Ok, let's use it
                        hi_bound_int = self._transformation.forward(hi_bound_ext)
                        delta_int = abs(hi_bound_int - self._get_internal_value())

                        break

                    else:

                        # Fix delta
                        self.delta = old_div(abs(self.value - self.min_value), 4.0)

                        if self.delta == 0:

                            # Parameter at the minimum
                            self.delta = old_div(abs(self.value - self.max_value), 4.0)

                        # Try again
                        continue

            assert delta_int is not None, "Bug"

            return delta_int

    # Define the property "prior"

    def _get_prior(self):

        return self._prior

    def _set_prior(self, prior):
        """Set prior for this parameter. The prior must be a function accepting the current value of the parameter
        as input and giving the probability density as output."""

        if prior is None:

            # Removing prior

            self._prior = None

        else:

            # Try and call the prior with the current value of the parameter
            try:

                _ = prior(self.value)

            except:

                log.exception(
                    "Could not call the provided prior. "
                    + "Is it a function accepting the current value of the parameter?"
                )
                raise NotCallableOrErrorInCall()

            try:

                prior.set_units(self.unit, u.dimensionless_unscaled)

            except AttributeError:

                log.exception(
                    "It looks like the provided prior is not a astromodels function."
                )

                raise NotCallableOrErrorInCall()

            self._prior = prior

    prior = property(
        _get_prior,
        _set_prior,
        doc="Gets or sets the current prior for this parameter. The prior must be a callable function "
        "accepting the current value of the parameter as input and returning the probability "
        "density as output. Set to None to remove prior.",
    )

    def has_prior(self):
        """
        Check whether the current parameter has a defined prior

        :return: True or False
        """

        return self._prior is not None

    def set_uninformative_prior(self, prior_class):
        """
        Sets the prior for the parameter to a uniform prior between the current minimum and maximum, or a
        log-uniform prior between the current minimum and maximum.

        NOTE: if the current minimum and maximum are not defined, the default bounds for the prior class will be used.

        :param prior_class : the class to be used as prior (either Log_uniform_prior or Uniform_prior, or a class which
        provide a lower_bound and an upper_bound properties)
        :return: (none)
        """

        prior_instance = prior_class()

        if self.min_value is None:

            raise ParameterMustHaveBounds(
                "Parameter %s does not have a defined minimum. Set one first, then re-run "
                "set_uninformative_prior" % self.path
            )

        else:

            try:

                prior_instance.lower_bound = self.min_value

            except SettingOutOfBounds:

                raise SettingOutOfBounds(
                    "Cannot use minimum of %s for prior %s"
                    % (self.min_value, prior_instance.name)
                )

        if self.max_value is None:

            raise ParameterMustHaveBounds(
                "Parameter %s does not have a defined maximum. Set one first, then re-run "
                "set_uninformative_prior" % self.path
            )

        else:  # pragma: no cover

            try:

                prior_instance.upper_bound = self.max_value

            except SettingOutOfBounds:

                raise SettingOutOfBounds(
                    "Cannot use maximum of %s for prior %s"
                    % (self.max_value, prior_instance.name)
                )

        assert np.isfinite(prior_instance.upper_bound.value), (
            "The parameter %s must have a finite maximum" % self.name
        )
        assert np.isfinite(prior_instance.lower_bound.value), (
            "The parameter %s must have a finite minimum" % self.name
        )

        self._set_prior(prior_instance)

    # Define property "free"

    def _set_free(self, value=True):

        self._free = value

    def _get_free(self):

        return self._free

    free = property(
        _get_free,
        _set_free,
        doc="Gets or sets whether the parameter is free or not. Use booleans, like: 'p.free = True' "
        " or 'p.free = False'. ",
    )

    # Define property "fix"

    def _set_fix(self, value=True):

        self._free = not value

    def _get_fix(self):

        return not self._free

    fix = property(
        _get_fix,
        _set_fix,
        doc="Gets or sets whether the parameter is fixed or not. Use booleans, like: 'p.fix = True' "
        " or 'p.fix = False'. ",
    )

    def add_auxiliary_variable(self, variable, law):

        # Assign units to the law
        law.set_units(variable.unit, self.unit)

        # Test the law
        try:

            _ = law(variable.value)

        except:  # pragma: no cover

            log.exception("The provided law for the auxiliary variable failed on call")

            raise NotCallableOrErrorInCall()

        self._aux_variable["law"] = law
        self._aux_variable["variable"] = variable

        # Now add the law as an attribute
        # so the user will be able to access its parameters as this.name.parameter_name

        # Now add the nodes

        if self._has_child(law.name):

            self._remove_child(law.name)

        self._add_child(law)

        # This parameter is not free anymore

        # First make a backup of the old status, so that it will be restored if the
        # law is removed
        self._old_free = self._free

        self.free = False

    def remove_auxiliary_variable(self):
        """
        Remove an existing auxiliary variable

        :return:
        """

        if not self.has_auxiliary_variable():

            # do nothing, but print a warning

            log.warning("Cannot remove a non-existing auxiliary variable")

        else:

            # Remove the law from the children

            self._remove_child(self._aux_variable["law"].name)

            # Clean up the dictionary

            self._aux_variable = {}

            # Set the parameter to the status it has before the auxiliary variable was created

            self.free = self._old_free

    def has_auxiliary_variable(self):
        """
        Returns whether the parameter is linked to an auxiliary variable
        """
        if self._aux_variable:

            return True

        else:

            return False

    @property
    def auxiliary_variable(self):
        """
        Returns a tuple with the auxiliary variable and the law

        :return: tuple (variable, law)
        """
        return self._aux_variable["variable"], self._aux_variable["law"]

    def _repr__base(self, rich_output=False):

        if not self.has_auxiliary_variable():

            representation = (
                "Parameter %s = %s [%s]\n"
                "(min_value = %s, max_value = %s, delta = %s, free = %s)"
                % (
                    self.name,
                    self.value,
                    self.unit,
                    self.min_value,
                    self.max_value,
                    self.delta,
                    self.free,
                )
            )

            if self._prior is not None:

                representation += " [prior: %s]" % self.prior.name

        else:

            representation = (
                "Parameter %s = %s [%s]\n"
                "(linked to auxiliary variable '%s' with law '%s')"
                % (
                    self.name,
                    self.value,
                    self.unit,
                    self._aux_variable["variable"].name,
                    self._aux_variable["law"].name,
                )
            )

        return representation

    def to_dict(self, minimal=False):

        """Returns the representation for serialization"""

        data = super(Parameter, self).to_dict()

        # Add wether is a normalization or not
        data["is_normalization"] = self._is_normalization

        if minimal:

            # No need to add anything
            pass

        else:

            # In the complete representation we output everything is needed to re-build the object

            if self.has_auxiliary_variable():

                # Store the function and the auxiliary variable

                data["value"] = "f(%s)" % self._aux_variable["variable"]._get_path()

                aux_variable_law_data = collections.OrderedDict()
                aux_variable_law_data[
                    self._aux_variable["law"].name
                ] = self._aux_variable["law"].to_dict()

                data["law"] = aux_variable_law_data

            # delta and free are attributes of Parameter, but not of ParameterBase

            data["delta"] = self._to_python_type(self._delta)
            data["free"] = self.free

            if self.has_prior():

                data["prior"] = {self.prior.name: self.prior.to_dict()}

        return data

    def get_randomized_value(self, variance=0.1):

        # Get a value close to the current value, but not identical
        # (used for the inizialization of Bayesian samplers)

        min_value = self.min_value
        max_value = self.max_value
        value = self.value

        if (min_value is not None) or (max_value is not None):

            # If _value is zero, then std will be zero, which doesn't make sense
            assert value != 0, (
                "You cannot randomize parameter %s because its value is exactly zero"
                % self.path
            )

            # Bounded parameter. Use a truncated normal so we are guaranteed
            # to have a random value within the boundaries

            std = abs(variance * value)

            if min_value is not None:

                a = old_div((min_value - value), std)

            else:

                a = -np.inf

            if max_value is not None:

                b = old_div((max_value - value), std)

            else:

                b = np.inf

            sample = scipy.stats.truncnorm.rvs(a, b, loc=value, scale=std, size=1)

            if (min_value is not None and sample < min_value) or (
                max_value is not None and sample > max_value
            ):  # pragma: no cover

                # This should never happen

                raise AssertionError(
                    "Got a sample outside of the boundaries of the truncated normal distribution"
                )

            return sample[0]

        else:

            # The parameter has no boundaries

            return np.random.normal(value, abs(variance * value))


class IndependentVariable(ParameterBase):
    """
    An independent variable like time or energy.
    """

    # Override the constructor to make the unit specification mandatory

    def __init__(self, name, value, unit, min_value=None, max_value=None, desc=None):

        super(IndependentVariable, self).__init__(
            name, value, unit=unit, min_value=min_value, max_value=max_value, desc=desc
        )

    def _repr__base(self, rich_output=False):

        return "IndependentVariable %s = %s\n" "(min_value = %s, max_value = %s)" % (
            self.name,
            self.value,
            self.min_value,
            self.max_value,
        )



from __future__ import division

import collections
from typing import Any, Dict, List, Optional, Union

import astropy.units as u
import numba as nb
import numpy
import scipy.integrate
from past.utils import old_div

from astromodels.core.memoization import use_astromodels_memoization
from astromodels.core.parameter import Parameter
from astromodels.core.sky_direction import SkyDirection
from astromodels.core.spectral_component import SpectralComponent
from astromodels.core.tree import Node
from astromodels.core.units import get_units
from astromodels.functions.function import Function1D
from astromodels.sources.source import Source, SourceType
from astromodels.utils.logging import setup_logger
from astromodels.utils.pretty_list import dict_to_list

__author__ = 'giacomov'

__all__ = ["PointSource"]

log = setup_logger(__name__)

class PointSource(Source, Node):
    """
    A point source. You can instance this class in many ways.

    - with Equatorial position and a function as spectrum (the component will be automatically called 'main')::

        >>> from astromodels import *
        >>> point_source = PointSource('my_source', 125.6, -75.3, Powerlaw())

    - with Galactic position and a function as spectrum (the component will be automatically called 'main')::

        >>> point_source = PointSource('my_source', l=15.67, b=80.75, spectral_shape=Powerlaw())

    - with Equatorial position or Galactic position and a list of spectral components::

        >>> c1 = SpectralComponent("component1", Powerlaw())
        >>> c2 = SpectralComponent("component2", Powerlaw())
        >>> point_source = PointSource("test_source",125.6, -75.3,components=[c1,c2])

        Or with Galactic position:

        >>> point_source = PointSource("test_source",l=15.67, b=80.75,components=[c1,c2])
    
    NOTE: by default the position of the source is fixed (i.e., its positional parameters are fixed)
    
    :param source_name: name for the source
    :param ra: Equatorial J2000 Right Ascension (ICRS)
    :param dec: Equatorial J2000 Declination (ICRS)
    :param spectral_shape: a 1d function representing the spectral shape of the source
    :param l: Galactic latitude
    :param b: Galactic longitude
    :param components: list of spectral components (instances of SpectralComponent)
    :param sky_position: an instance of SkyDirection
    :return:
    """
    def __init__(self,
                 source_name: str,
                 ra: Optional[float] = None,
                 dec: Optional[float] = None,
                 spectral_shape: Optional[Function1D] = None,
                 l: Optional[float] = None,
                 b: Optional[float] = None,
                 components=None,
                 sky_position: Optional[SkyDirection]=None):

        # Check that we have all the required information

        # (the '^' operator acts as XOR on booleans)

        # Check that we have one and only one specification of the position

        if not ((ra is not None and dec is not None) ^
                (l is not None and b is not None) ^
                (sky_position is not None)):

            log.error(
                "You have to provide one and only one specification for the position"
            )

            raise AssertionError()

        # Gather the position

        if not isinstance(sky_position, SkyDirection):

            if (ra is not None) and (dec is not None):

                # Check that ra and dec are actually numbers

                try:

                    ra = float(ra)
                    dec = float(dec)

                except (TypeError, ValueError):

                    log.error("RA and Dec must be numbers. If you are confused by this message, you "
                                         "are likely using the constructor in the wrong way. Check the documentation.")
                    
                    raise AssertionError()

                sky_position = SkyDirection(ra=ra, dec=dec)

            else:

                sky_position = SkyDirection(l=l, b=b)

        self._sky_position: SkyDirection = sky_position

        # Now gather the component(s)

        # We need either a single component, or a list of components, but not both
        # (that's the ^ symbol)

        if not (spectral_shape is not None) ^ (components is not None):

            log.error("You have to provide either a single component, or a list of components (but not both).")

            raise AssertionError()

        # If the user specified only one component, make a list of one element with a default name ("main")

        if spectral_shape is not None:

            components = [SpectralComponent("main", spectral_shape)]

        Source.__init__(self, components, src_type=SourceType.POINT_SOURCE)

        # A source is also a Node in the tree

        Node.__init__(self, source_name)

        # Add the position as a child node, with an explicit name

        self._add_child(self._sky_position)

        # Add a node called 'spectrum'

        spectrum_node = Node('spectrum')
        spectrum_node._add_children(list(self._components.values()))

        self._add_child(spectrum_node)

        # Now set the units
        # Now sets the units of the parameters for the energy domain

        current_units = get_units()

        # Components in this case have energy as x and differential flux as y

        x_unit = current_units.energy
        y_unit = (current_units.energy * current_units.area * current_units.time) ** (-1)

        # Now set the units of the components
        for component in list(self._components.values()):

            component.shape.set_units(x_unit, y_unit)

    def __call__(self, x, tag=None):

        if tag is None:

            # No integration nor time-varying or whatever-varying

            if isinstance(x, u.Quantity):

                # Slow version with units

                results = [component.shape(x) for component in list(self.components.values())]

                # We need to sum like this (slower) because using np.sum will not preserve the units
                # (thanks astropy.units)

                return sum(results)

            else:

                # Fast version without units, where x is supposed to be in the same units as currently defined in
                # units.get_units()

                results = numpy.array([component.shape(x) for component in list(self.components.values())])

                return _sum(results)

        else:

            # Time-varying or energy-varying or whatever-varying

            integration_variable, a, b = tag

            if b is None:

                # Evaluate in a, do not integrate

                # Suspend memoization because the memoization gets confused when integrating
                with use_astromodels_memoization(False):

                    integration_variable.value = a

                    res = self.__call__(x, tag=None)

                return res

            else:

                # Integrate between a and b

                integrals = numpy.zeros(len(x))

                # TODO: implement an integration scheme avoiding the for loop

                # Suspend memoization because the memoization gets confused when integrating
                with use_astromodels_memoization(False):

                    reentrant_call = self.__call__

                    for i, e in enumerate(x):

                        def integral(y):

                            integration_variable.value = y

                            return reentrant_call(e, tag=None)

                        # Now integrate
                        integrals[i] = scipy.integrate.quad(integral, a, b, epsrel=1e-5)[0]

                return old_div(integrals, (b - a))

    def has_free_parameters(self) -> bool:
        """
        Returns True or False whether there is any parameter in this source

        :return:
        """

        for component in list(self._components.values()):

            for par in list(component.shape.parameters.values()):

                if par.free:

                    return True

        for par in list(self.position.parameters.values()):

            if par.free:

                return True

        return False

    @property
    def free_parameters(self) -> Dict[str, Parameter]:
        """
        Returns a dictionary of free parameters for this source.
        We use the parameter path as the key because it's 
        guaranteed to be unique, unlike the parameter name.

        :return:
        """
        free_parameters = collections.OrderedDict()

        for component in list(self._components.values()):

            for par in list(component.shape.parameters.values()):

                if par.free:

                    free_parameters[par.path] = par

        for par in list(self.position.parameters.values()):

            if par.free:

                free_parameters[par.path] = par

        return free_parameters

    @property
    def parameters(self) -> Dict[str, Parameter]:
        """
        Returns a dictionary of all parameters for this source.
        We use the parameter path as the key because it's 
        guaranteed to be unique, unlike the parameter name.

        :return:
        """
        all_parameters = collections.OrderedDict()

        for component in self._components.values():

            for par in component.shape.parameters.values():

                all_parameters[par.path] = par

        for par in self.position.parameters.values():

            all_parameters[par.path] = par

        return all_parameters

    def _repr__base(self, rich_output=False):
        """
        Representation of the object

        :param rich_output: if True, generates HTML, otherwise text
        :return: the representation
        """

        # Make a dictionary which will then be transformed in a list

        repr_dict = collections.OrderedDict()

        key = '%s (point source)' % self.name

        repr_dict[key] = collections.OrderedDict()
        repr_dict[key]['position'] = self._sky_position.to_dict(minimal=True)
        repr_dict[key]['spectrum'] = collections.OrderedDict()

        for component_name, component in list(self.components.items()):

            repr_dict[key]['spectrum'][component_name] = component.to_dict(minimal=True)

        return dict_to_list(repr_dict, rich_output)



@nb.njit(fastmath=True)
def _sum(x):
    return numpy.sum(x, axis=0)
    



from __future__ import division, print_function

import os
import shutil
from pathlib import Path

import numpy as np
import numpy.testing as npt
import pytest

from astromodels import Model, PointSource, clone_model, load_model
from astromodels.functions import (Band, MissingDataFile, Powerlaw,
                                   TemplateModel, TemplateModelFactory,
                                   XSPECTableModel)
from astromodels.functions.template_model import convert_old_table_model
from astromodels.utils.data_files import _get_data_file_path
from astromodels.utils.logging import update_logging_level

update_logging_level("DEBUG")


import pickle

__author__ = 'giacomov'


def get_comparison_function():

    mo = Band()
    mo.K = 1

    return mo


@pytest.mark.slow
def test_template_factory_1D():

    mo = get_comparison_function()

    energies = np.logspace(1, 3, 50)

    t = TemplateModelFactory('__test1D', 'A test template', energies, ['alpha'])

    alpha_grid = np.linspace(-1.5, 1, 15)
    #beta_grid = np.linspace(-3.5, -1.6, 15)
    #xp_grid = np.logspace(1, 3, 20)

    t.define_parameter_grid('alpha', alpha_grid)


    for a in alpha_grid:
        mo.alpha = a
        mo.beta = -2.5
        mo.xp = 300.


        t.add_interpolation_data(mo(energies), alpha=a)

    print("Data has been prepared")

    t.save_data(overwrite=True)

@pytest.mark.slow
def test_template_factory():

    mo = get_comparison_function()

    energies = np.logspace(1, 3, 50)

    t = TemplateModelFactory('__test', 'A test template', energies, ['alpha', 'xp', 'beta'])

    alpha_grid = np.linspace(-1.5, 1, 15)
    beta_grid = np.linspace(-3.5, -1.6, 15)
    xp_grid = np.logspace(1, 3, 20)

    t.define_parameter_grid('alpha', alpha_grid)
    t.define_parameter_grid('beta', beta_grid)
    t.define_parameter_grid('xp', xp_grid)

    for a in alpha_grid:

        for b in beta_grid:

            for xp in xp_grid:
                mo.alpha = a
                mo.beta = b
                mo.xp = xp

                t.add_interpolation_data(mo(energies), alpha=a, xp=xp, beta=b)

    print("Data has been prepared")

    t.save_data(overwrite=True)

    tm = TemplateModel('__test')

    tm(energies)

    tm.clean()
    



# This will be run second, so the template will exist
@pytest.mark.slow
def test_template_function():

    tm = TemplateModel('__test')

    mo = get_comparison_function()

    new_alpha_grid = np.linspace(-1.5, 1, 15)
    new_beta_grid = np.linspace(-3.5, -1.6, 15)
    new_xp_grid = np.logspace(1, 3, 15)

    new_energies = np.logspace(1, 3, 40)

    tm.K = 1

    mo.K = 1

    for a in new_alpha_grid:

        for b in new_beta_grid:

            for xp in new_xp_grid:

                mo.alpha = a
                mo.beta = b
                mo.xp = xp

                tm.alpha = a
                tm.beta = b
                tm.xp = xp

                res1 = mo(new_energies)
                res2 = tm(new_energies)

                deltas = np.abs((res2 - res1) / res1)

                idx = deltas > 0.1

                if np.any(idx):

                    raise AssertionError("Interpolation precision @ %s is %s, "
                                         "worse than 10 percent, "
                                         "with parameters %s!" % (new_energies[idx], deltas[idx], [a,b,xp]))

@pytest.mark.slow
def test_input_output():

    tm = TemplateModel('__test')
    tm.alpha = -0.95
    tm.beta = -2.23

    fake_source = PointSource("test", ra=0.0, dec=0.0, spectral_shape=tm)

    fake_model = Model(fake_source)

    clone = clone_model(fake_model)

    assert clone.get_number_of_point_sources() == 1
    assert tm.data_file == clone.test.spectrum.main.shape.data_file

    assert clone.test.spectrum.main.shape.alpha.value == tm.alpha.value
    assert clone.test.spectrum.main.shape.beta.value == tm.beta.value

    xx = np.linspace(1, 10, 100)

    assert np.allclose(clone.test.spectrum.main.shape(xx), fake_model.test.spectrum.main.shape(xx))

    # Test pickling
    dump = pickle.dumps(clone)

    clone2 = pickle.loads(dump)

    assert clone2.get_number_of_point_sources() == 1
    assert tm.data_file == clone2.test.spectrum.main.shape.data_file
    assert np.allclose(clone2.test.spectrum.main.shape(xx), fake_model.test.spectrum.main.shape(xx))

    # Test pickling with other functions
    new_shape = tm * Powerlaw()

    new_shape.index_2 = -2.256

    dump2 = pickle.dumps(new_shape)

    clone3 = pickle.loads(dump2)

    assert clone3.index_2.value == new_shape.index_2.value

    # Now save to disk and reload
    fake_source2 = PointSource("test", ra=0.0, dec=0.0, spectral_shape=new_shape)

    fake_model2 = Model(fake_source2)

    fake_model2.save("__test.yml", overwrite=True)

    # Now try to reload
    reloaded_model = load_model("__test.yml")

    assert reloaded_model.get_number_of_point_sources() == 1
    assert np.allclose(fake_model2.test.spectrum.main.shape(xx), reloaded_model.test.spectrum.main.shape(xx))

    os.remove("__test.yml")

def test_xspec_table_model():

    test_table = _get_data_file_path("tests/test_xspec_table_model.fits")

    xtm = XSPECTableModel(test_table)

    xtm.to_table_model('xspectm_test', 'xspec model', overwrite=True)



def test_table_conversion():

    old_table_file = _get_data_file_path("tests/old_table.h5")

    p = Path.home() / ".astromodels" / "data" / "old_table.h5"
    
    
    shutil.copy(old_table_file, p)

    # convert the table

    convert_old_table_model("old_table")

    # now load the old table

    old_table = TemplateModel("old_table")


    # should be the same as in test

    test = TemplateModel("__test")

    xx = np.logspace(1, 3, 50)

    
    npt.assert_almost_equal(test(xx), old_table(xx))
    
    p.unlink()
    
    

    
    
    



from __future__ import absolute_import

import os

from ._version import get_versions

# Import the version

#
#

if os.environ.get("ASTROMODELS_DEBUG", None) is None:

    from .utils.configuration import astromodels_config
    from .core.memoization import use_astromodels_memoization
    from .core.model import Model
    from .core.model_parser import clone_model, load_model
    from .core.parameter import (IndependentVariable, Parameter,
                                 SettingOutOfBounds)
    from .core.polarization import LinearPolarization, StokesPolarization
    from .core.serialization import *
    from .core.spectral_component import SpectralComponent
    from .core.units import get_units
    from .functions import (
        Band, Band_Calderone, Band_grbm, Broken_powerlaw, Cutoff_powerlaw,
        Inverse_cutoff_powerlaw, Powerlaw, Powerlaw_Eflux, Powerlaw_flux,
        SmoothlyBrokenPowerLaw, Super_cutoff_powerlaw, Constant, Cubic,
        DiracDelta, Exponential_cutoff, Line, Quadratic, Sin, StepFunction,
        StepFunctionUpper, PhAbs, TbAbs, WAbs, Asymm_Gaussian_on_sphere,
        Disk_on_sphere, Ellipse_on_sphere, Gaussian_on_sphere,
        Latitude_galactic_diffuse, Power_law_on_sphere, SpatialTemplate_2D,
        Continuous_injection_diffusion, Continuous_injection_diffusion_ellipse,
        Continuous_injection_diffusion_legacy, GalPropTemplate_3D, DMSpectra,
        DMFitFunction, Cauchy, Cosine_Prior, Gaussian, Log_normal, Quartic,
        get_polynomial, Log_uniform_prior, Truncated_gaussian, Uniform_prior,
        TemplateModel, TemplateModelFactory, XSPECTableModel, MissingDataFile,
        Log_parabola, Blackbody, Function1D, Function2D, Function3D,
        FunctionMeta, ModelAssertionViolation, has_naima, has_gsl,
        has_ebltable, has_atomdb)

    if has_ebltable:

        from .functions import EBLattenuation

    if has_gsl:

        from .functions import Cutoff_powerlaw_flux

    if has_naima:

        from .functions import Synchrotron

    if has_atomdb:

        from .functions import APEC, VAPEC
        
    from .functions.function import get_function_class, list_functions
    from .sources import ExtendedSource, PointSource, ParticleSource


    astromodels_units = get_units()
    from astromodels.utils.logging import setup_logger, update_logging_level, silence_warnings, activate_warnings

import astropy.units as u

log = setup_logger(__name__)

__version__ = get_versions()['version']
del get_versions



import os
import re
import sys

# adapted from thepipe
# https://github.com/tamasgal/thepipe/


ATTRIBUTES = dict(
    list(
        zip([
            'bold', 'dark', '', 'underline', 'blink', '', 'reverse',
            'concealed'
        ], list(range(1, 9)))))
del ATTRIBUTES['']

ATTRIBUTES_RE = r'\033\[(?:%s)m' % '|'  \
                .join(['%d' % v for v in ATTRIBUTES.values()])

HIGHLIGHTS = dict(
    list(
        zip([
            'on_grey', 'on_red', 'on_green', 'on_yellow', 'on_blue',
            'on_magenta', 'on_cyan', 'on_white'
        ], list(range(40, 48)))))

HIGHLIGHTS_RE = r'\033\[(?:%s)m' % '|'  \
                .join(['%d' % v for v in HIGHLIGHTS.values()])

COLORS = dict(
    list(
        zip([
            'grey',
            'red',
            'green',
            'yellow',
            'blue',
            'magenta',
            'cyan',
            'white',
        ], list(range(30, 38)))))

COLORS_RE = r'\033\[(?:%s)m' % '|'.join(['%d' % v for v in COLORS.values()])

RESET = r'\033[0m'
RESET_RE = r'\033\[0m'


def colored(text, color=None, on_color=None, attrs=None, ansi_code=None):
    """Colorize text, while stripping nested ANSI color sequences.
    Author:  Konstantin Lepa <konstantin.lepa@gmail.com> / termcolor
    Available text colors:
        red, green, yellow, blue, magenta, cyan, white.
    Available text highlights:
        on_red, on_green, on_yellow, on_blue, on_magenta, on_cyan, on_white.
    Available attributes:
        bold, dark, underline, blink, reverse, concealed.
    Example:
        colored('Hello, World!', 'red', 'on_grey', ['blue', 'blink'])
        colored('Hello, World!', 'green')
    """
    if os.getenv('ANSI_COLORS_DISABLED') is None:
        if ansi_code is not None:
            return "\033[38;5;{}m{}\033[0m".format(ansi_code, text)
        fmt_str = '\033[%dm%s'
        if color is not None:
            text = re.sub(COLORS_RE + '(.*?)' + RESET_RE, r'\1', text)
            text = fmt_str % (COLORS[color], text)
        if on_color is not None:
            text = re.sub(HIGHLIGHTS_RE + '(.*?)' + RESET_RE, r'\1', text)
            text = fmt_str % (HIGHLIGHTS[on_color], text)
        if attrs is not None:
            text = re.sub(ATTRIBUTES_RE + '(.*?)' + RESET_RE, r'\1', text)
            for attr in attrs:
                text = fmt_str % (ATTRIBUTES[attr], text)
        return text + RESET
    else:
        return text


def cprint(text, color=None, on_color=None, attrs=None):
    """Print colorize text.
    Author:  Konstantin Lepa <konstantin.lepa@gmail.com> / termcolor
    It accepts arguments of print function.
    """
    print((colored(text, color, on_color, attrs)))


def isnotebook():
    """Check if running within a Jupyter notebook"""
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False


def supports_color():
    """Checks if the terminal supports color."""
    if isnotebook():
        return True
    supported_platform = sys.platform != 'win32' or 'ANSICON' in os.environ
    is_a_tty = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()

    if not supported_platform or not is_a_tty:
        return False

    return True



import logging
from dataclasses import dataclass
from enum import IntEnum, Enum


# from .catalog_structure import Catalogs, PublicDataServer
# from .fitting_structure import BayesianDefault, MLEDefault
# from .plotting_structure import GenericPlotting, ModelPlotting
# from .plugin_structure import Plugins, TimeSeries
# from .point_source_structure import PointSourceDefaults

# logging
class LoggingLevel(IntEnum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


@dataclass
class Logging:

    path: str = "~/.astromodels/log"
    developer: bool = 'off'
    usr: bool = 'on'
    console: bool = 'on'
    level: LoggingLevel = LoggingLevel.INFO
    startup_warnings: bool = 'on'


class AbsTables(Enum):
    WILM = "WILM"
    ASPL = "ASPL"
    AG89 = "AG89" 


class EBLTable(Enum):
    franceschini = "franceschini"
    kneiske = "kneiske"
    dominguez = "dominguez"
    inuoe = "inuoe"
    gilmore = "gilmore"

    
    
@dataclass
class AbsorptionModels:
    tbabs_table: AbsTables = AbsTables.WILM
    phabs_table: AbsTables = AbsTables.AG89
    ebl_table: EBLTable = EBLTable.dominguez


    

@dataclass
class Config:
    logging: Logging = Logging()
    absorption_models: AbsorptionModels = AbsorptionModels()



from builtins import range
import collections
import functools
import contextlib
import astropy.units as u

_WITH_MEMOIZATION = False
_CACHE_SIZE = 20


@contextlib.contextmanager
def use_astromodels_memoization(switch, cache_size=_CACHE_SIZE):
    """
    Activate/deactivate memoization temporarily

    :param switch: True (memoization on) or False (memoization off)
    :param cache_size: number of previous evaluation of functions to keep in memory. Default: 100
    :return:
    """

    global _WITH_MEMOIZATION
    global _CACHE_SIZE

    old_status = bool(_WITH_MEMOIZATION)
    old_cache_size = int(_CACHE_SIZE)

    _WITH_MEMOIZATION = bool(switch)
    _CACHE_SIZE = int(cache_size)

    yield

    _WITH_MEMOIZATION = old_status
    _CACHE_SIZE = old_cache_size



def memoize(method):
    """
    A decorator for functions of sources which memoize the results of the last _CACHE_SIZE calls,

    :param method: method to be memoized
    :return: the decorated method
    """

    cache = method.cache = collections.OrderedDict()

    # Put these two methods in the local space (faster)
    _get = cache.get
    _popitem = cache.popitem

    @functools.wraps(method)
    def memoizer(instance, x, *args, **kwargs):

        if not _WITH_MEMOIZATION or isinstance(x, u.Quantity):

            # Memoization is not active or using units, do not use memoization

            return method(instance, x, *args, **kwargs)

        # Create a tuple because a tuple is hashable

        unique_id = tuple(float(yy.value) for yy in list(instance.parameters.values())) + (x.size, x.min(), x.max())

        # Create a unique identifier for this combination of inputs

        key = hash(unique_id)

        # Let's do it this way so we only look into the dictionary once

        result = _get(key)

        if result is not None:

            return result

        else:

            result = method(instance, x, *args, **kwargs)

            cache[key] = result

            if len(cache) > _CACHE_SIZE:
                # Remove half of the element (but at least 1, even if _CACHE_SIZE=1, which would be pretty idiotic ;-) )
                [_popitem(False) for i in range(max(_CACHE_SIZE // 2, 1))]

            return result

    # Add the function as a "attribute" so we can access it
    memoizer.input_object = method

    return memoizer



from future import standard_library

standard_library.install_aliases()
import collections
import os
import uuid
from builtins import object

from astromodels.core.node_ctype import _Node
from astromodels.utils.io import display
from astromodels.utils.valid_variable import is_valid_variable_name

from .cpickle_compatibility_layer import cPickle


class DuplicatedNode(Exception):
    pass


class ProtectedAttribute(RuntimeError):
    pass


class NonExistingAttribute(RuntimeWarning):
    pass


# This is necessary for pickle to be able to reconstruct a NewNode class (or derivate)
# during unpickling
class NewNodeUnpickler(object):

    def __call__(self, cls):

        instance = cls.__new__(cls)

        return instance


class Node(_Node):

    # This apparently dumb constructor is needed otherwise pickle will fail

    def __init__(self, name):

        assert is_valid_variable_name(name), "Illegal characters in name %s. You can only use letters and numbers, " \
                                             "and _" % name

        assert name != "name", "You cannot call a node 'name', it is reserved."

        _Node.__init__(self, name)
        
    #########################################################################
    # The next 3 methods are *really* necessary for anything to work

    def __reduce__(self):

        state = {}
        state['children'] = self._get_children()
        state['name'] = self.name
        state['__dict__'] = self.__dict__

        return NewNodeUnpickler(), (self.__class__,), state

    def __setstate__(self, state):

        # Set the children

        self._add_children(state['children'])

        # Set the name of this node

        self._change_name(state['name'])

        # Restore everything else
        
        for k in state['__dict__']:

            self.__dict__[k] = state['__dict__'][k]

    # This is necessary for copy.deepcopy to work
    def __deepcopy__(self, memodict={}):

        return cPickle.loads(cPickle.dumps(self))

    #########################################################################

    # This is used by dir() and by the autocompletion in Ipython
    def __dir__(self):

        # Get the names of the attributes of the class
        l = list(self.__class__.__dict__.keys())

        # Get all the children
        l.extend([child.name for child in self._get_children()])

        return l

    def to_dict(self, minimal=False):

        this_dict = collections.OrderedDict()

        for child in self._get_children():

            this_dict[child.name] = child.to_dict(minimal)

        return this_dict

    def _repr__base(self, rich_output):

        return "Node with name %s" % self.name

    def __repr__(self):
        """
        Textual representation for console

        :return: representation
        """

        return self._repr__base(rich_output=False)

    def _repr_html_(self):
        """
        HTML representation for the IPython notebook

        :return: HTML representation
        """

        return self._repr__base(rich_output=True)

    def display(self):
        """
        Display information about the point source.

        :return: (none)
        """

        # This will automatically choose the best representation among repr and repr_html

        display(self)


class OldNode(object):

    def __init__(self, name):

        self._children = collections.OrderedDict()
        self._parent = None

        assert is_valid_variable_name(name), "Illegal characters in name %s. You can only use letters and numbers, " \
                                             "and _" % name

        self._name = name

    @property
    def name(self):
        """
        Returns the name of the node

        :return: a string containing the name
        """
        return self._name

    @property
    def path(self):

        return ".".join(self._get_path())

    def _add_children(self, children):

        for child in children:

            self._add_child(child)

    def _add_child(self, new_child, name=None, add_attribute=True):

        new_child._set_parent(self)

        if name is None:

            name = new_child.name

        if name in self._children:

            raise DuplicatedNode("You cannot use the same name (%s) for different nodes" % name)

        self._children[name] = new_child

        if add_attribute:

            # Add also an attribute with the name of the new child, to allow access with a syntax like
            # node.child

            object.__setattr__(self, name, new_child)

    def _get_child(self, child_name):

        return self._children[child_name]

    def _remove_child(self, child_name):

        object.__delattr__(self, child_name)

        return self._children.pop(child_name)

    def _set_parent(self, parent):

        # We need to use directly the __setattr__ method because the normal self._children = ... will trigger
        # an exception, because of the __setattr__ method of the DualAccessClass which forbids changing
        # nodes. However, this might reassign the parent to a class

        object.__setattr__(self, "_parent", parent)

    def _get_parent(self):

        return self._parent

    def _get_child_from_path(self, path):
        """
        Return a children below this level, starting from a path of the kind "this_level.something.something.name"

        :param path: the key
        :return: the child
        """

        keys = path.split(".")

        this_child = self

        for key in keys:

            try:

                this_child = this_child._get_child(key)

            except KeyError:

                raise KeyError("Child %s not found" % path)

        return this_child

    def _get_path(self):

        parent_names = []

        current_node = self

        while True:

            this_parent = current_node._get_parent()

            if this_parent is None:

                break

            else:

                parent_names.append(current_node.name)

                current_node = this_parent

        return parent_names[::-1]

    def to_dict(self, minimal=False):

        this_dict = collections.OrderedDict()

        for key, val in list(self._children.items()):

            this_dict[key] = val.to_dict(minimal)

        return this_dict

    def _repr__base(self, rich_output):

        raise NotImplementedError("You should implement the __repr__base method for each class")

    def __repr__(self):
        """
        Textual representation for console

        :return: representation
        """

        return self._repr__base(rich_output=False)

    def _repr_html_(self):
        """
        HTML representation for the IPython notebook

        :return: HTML representation
        """

        return self._repr__base(rich_output=True)

    def display(self):
        """
        Display information about the point source.

        :return: (none)
        """

        # This will automatically choose the best representation among repr and repr_html

        display(self)

    def _find_instances(self, cls):
        """
        Find all the instances of cls below this node.

        :return: a dictionary of instances of cls
        """

        instances = collections.OrderedDict()

        for child_name, child in list(self._children.items()):

            if isinstance(child, cls):

                key_name = ".".join(child._get_path())

                instances[key_name] = child

                # Now check if the instance has children,
                # and if it does go deeper in the tree

                # NOTE: an empty dictionary evaluate as False

                if child._children:

                    instances.update(child._find_instances(cls))

            else:

                instances.update(child._find_instances(cls))

        return instances




# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

from __future__ import absolute_import
import json

version_json = '''
{
 "date": "2021-08-17T09:48:13+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "d4d4b0b2857d3fda4c177ac2066065f04a8877fc",
 "version": "2.2.1"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



from __future__ import print_function

import astropy.io.fits as fits
import astropy.units as u
import numpy as np
import pytest

from astromodels.core.model import Model
from astromodels.core.model_parser import clone_model
from astromodels.core.spectral_component import SpectralComponent
from astromodels.functions import *
from astromodels.functions import Log_parabola, Powerlaw
from astromodels.functions.function import _known_functions
from astromodels.sources.extended_source import ExtendedSource

__author__ = 'henrikef'


def make_test_template(ra, dec, fitsfile):

    # Test template function: 40 pixel (0.8 deg) wide square centered approximately around a given ra, dec.

    cards = {
        "SIMPLE": "T",
        "BITPIX": -32,
        "NAXIS": 2,
        "NAXIS1": 200,
        "NAXIS2": 200,
        "DATE": '2018-11-13',
        "CUNIT1": 'deg',
        "CRVAL1":  ra,
        "CRPIX1": 100,
        "CDELT1": -0.02,
        "CUNIT2": 'deg',
        "CRVAL2": dec,
        "CRPIX2": 100,
        "CDELT2": 0.02,
        "CTYPE1": 'RA---TAN',
        "CTYPE2": 'DEC--TAN'}

    data = np.zeros([200, 200])
    data[80:120, 80:120] = 1

    total = np.sum(data)
    dOmega = (abs(cards["CDELT1"] * cards["CDELT2"]) *
              u.degree*u.degree).to(u.steradian).value
    data = data / total / dOmega

    hdu = fits.PrimaryHDU(data=data, header=fits.Header(cards))
    hdu.writeto(fitsfile, overwrite=True)


def test_constructor():

    # RA, Dec and L,B of the same point in the sky

    ra, dec = (125.6, -75.3)
    l, b = (288.44190139183564, -20.717313145391525)

    # This should throw an error as we are using Powerlaw instead of Powerlaw()
    with pytest.raises(RuntimeError):

        _ = ExtendedSource("my_source", Gaussian_on_sphere, Powerlaw)

    # This should throw an error because we should use a 2D function for the spatial shape
    with pytest.raises(RuntimeError):

        _ = ExtendedSource("my_source", Powerlaw(), Powerlaw())

    # Init with RA, Dec

    shape = Gaussian_on_sphere()
    source1 = ExtendedSource('my_source', shape, Powerlaw())
    shape.lon0 = ra*u.degree
    shape.lat0 = dec*u.degree

    assert source1.spatial_shape.lon0.value == ra
    assert source1.spatial_shape.lat0.value == dec

    # Verify that the position is free by default
    assert source1.spatial_shape.lon0.free
    assert source1.spatial_shape.lon0.free


def test_call():

    # Multi-component

    po1 = Powerlaw()
    po2 = Powerlaw()

    c1 = SpectralComponent("component1", po1)
    c2 = SpectralComponent("component2", po2)

    ra, dec = (125.6, -75.3)

    def test_one(class_type, name):

        print("testing %s ..." % name)

        shape = class_type()
        source = ExtendedSource('test_source_%s' %
                                name, shape, components=[c1, c2])

        if name != "SpatialTemplate_2D":
            shape.lon0 = ra*u.degree
            shape.lat0 = dec*u.degree

        else:
            make_test_template(ra, dec, "__test.fits")
            shape.load_file("__test.fits")
            shape.K = 1.0

        assert np.all(source.spectrum.component1([1, 2, 3]) == po1([1, 2, 3]))
        assert np.all(source.spectrum.component2([1, 2, 3]) == po2([1, 2, 3]))

        one = source.spectrum.component1([1, 2, 3])
        two = source.spectrum.component2([1, 2, 3])

        # check spectral components
        assert np.all(
            np.abs(one + two - source.get_spatially_integrated_flux([1, 2, 3])) == 0)

        # check spectral and spatial components
        total = source([ra, ra, ra], [dec, dec, dec], [1, 2, 3])
        spectrum = one + two
        spatial = source.spatial_shape([ra, ra, ra], [dec, dec, dec])
        assert np.all(np.abs(total - spectrum*spatial) == 0)

        total = source([ra*1.01]*3, [dec*1.01]*3, [1, 2, 3])
        spectrum = one + two
        spatial = source.spatial_shape([ra*1.01]*3, [dec*1.01]*3)
        assert np.all(np.abs(total - spectrum*spatial) == 0)

    for key in _known_functions:

        if key in ["Latitude_galactic_diffuse"]:
            # not testing latitude galactic diffuse for now.
            continue

        this_function = _known_functions[key]

        if this_function._n_dim == 2 and not this_function().is_prior:

            test_one(this_function, key)

    with pytest.raises(AssertionError):
        # this will fail because the Latitude_galactic_diffuse function isn't normalized.
        test_one(
            _known_functions["Latitude_galactic_diffuse"], "Latitude_galactic_diffuse")


def test_call_with_units():

    # Multi-component

    po1 = Powerlaw()
    po2 = Powerlaw()

    c1 = SpectralComponent("component1", po1)
    c2 = SpectralComponent("component2", po2)

    ra, dec = (125.6, -75.3)

    def test_one(class_type, name):

        print("testing %s ..." % name)

        shape = class_type()
        source = ExtendedSource('test_source_%s' %
                                name, spatial_shape=shape, components=[c1, c2])

        if name != "SpatialTemplate_2D":
            shape.lon0 = ra*u.degree
            shape.lat0 = dec*u.degree

        else:
            make_test_template(ra, dec, "__test.fits")
            shape.load_file("__test.fits")
            shape.K = 1.0

        assert np.all(source.spectrum.component1(
            [1, 2, 3]*u.keV) == po1([1, 2, 3]*u.keV))
        assert np.all(source.spectrum.component2(
            [1, 2, 3]*u.keV) == po2([1, 2, 3]*u.keV))

        one = source.spectrum.component1([1, 2, 3]*u.keV)
        two = source.spectrum.component2([1, 2, 3]*u.keV)

        # check spectral components
        assert np.all(
            np.abs(one + two - source.get_spatially_integrated_flux([1, 2, 3]*u.keV)) == 0)

        # check spectral and spatial components
        #spatial = source.spatial_shape( ra*u.deg,dec*u.deg )
        spatial = source.spatial_shape(
            [ra, ra, ra]*u.deg, [dec, dec, dec]*u.deg)

        total = source([ra, ra, ra]*u.deg, [dec, dec, dec]
                       * u.deg, [1, 2, 3]*u.keV)
        spectrum = one + two
        assert np.all(np.abs(total - spectrum*spatial) == 0)

        total = source([ra*1.01]*3*u.deg, [dec*1.01]*3*u.deg, [1, 2, 3]*u.keV)
        spectrum = one + two
        spatial = source.spatial_shape([ra*1.01]*3*u.deg, [dec*1.01]*3*u.deg)
        assert np.all(np.abs(total - spectrum*spatial) == 0)

        model = Model(source)
        new_model = clone_model(model)

        new_total = new_model['test_source_%s' % name](
            [ra*1.01]*3*u.deg, [dec*1.01]*3*u.deg, [1, 2, 3]*u.keV)
        assert np.all(np.abs(total - new_total) == 0)

    for key in _known_functions:

        if key in ["Latitude_galactic_diffuse"]:
            # not testing latitude galactic diffuse for now.
            continue

        this_function = _known_functions[key]

        if this_function._n_dim == 2 and not this_function().is_prior:

            test_one(this_function, key)

    with pytest.raises(AssertionError):
        # this will fail because the Latitude_galactic_diffuse function isn't normalized.
        test_one(
            _known_functions["Latitude_galactic_diffuse"], "Latitude_galactic_diffuse")


def test_free_param():

    spectrum = Log_parabola()
    source = ExtendedSource(
        "test_source", spatial_shape=Gaussian_on_sphere(), spectral_shape=spectrum)

    parameters = [spectrum.alpha, spectrum.beta, spectrum.piv, spectrum.K,
                  source.spatial_shape.lat0, source.spatial_shape.lon0, source.spatial_shape.sigma]

    for param in parameters:
        param.free = False

    assert len(source.free_parameters) == 0

    for i, param in enumerate(parameters):
        param.free = True
        assert len(source.free_parameters) == i+1



from future import standard_library
standard_library.install_aliases()
from astromodels.core.model import Model
from astromodels.core.model_parser import ModelParser

# We do not want to actually import anything from this module
__all__ = []

# copyreg is called copy_reg in python2
try:

    import copyreg #py3

except ImportError:

    import copyreg as copyreg #py2


# Model serializer
def serialize_model(model):

    dict_with_types = model.to_dict_with_types()

    return unserialize_model, (dict_with_types,)


def unserialize_model(dict_with_types):

    return ModelParser(model_dict=dict_with_types).get_model()

# Register serialization/unserialization for Model
copyreg.constructor(unserialize_model)
copyreg.pickle(Model, serialize_model, unserialize_model)



from __future__ import print_function
from __future__ import division
import astropy.units as u
import numpy as np
import numpy.testing as npt
import pytest

from astromodels.core.spectral_component import SpectralComponent
from astromodels.functions import Powerlaw, Exponential_cutoff, Log_parabola, Blackbody, Band
try:
    from astromodels.functions import TbAbs, PhAbs, WAbs

    has_abs_models = True

except:

    has_abs_models = False

    
from astromodels.sources.point_source import PointSource
from astromodels.sources.particle_source import ParticleSource
from astromodels.core.model import Model
from astromodels.core.model_parser import clone_model, load_model



try:

    from astromodels.xspec import XS_phabs, XS_powerlaw

except:

    has_xspec = False

else:

    has_xspec = True

try:

    from astromodels.functions import EBLattenuation

except:

    has_ebl = False

else:

    has_ebl = True

from astromodels.functions.priors import *
from astromodels.functions.function import _known_functions

__author__ = 'giacomov'



_multiplicative_models = ["PhAbs", "TbAbs", "WAbs", "EBLattenuation"]

def test_constructor():

    # RA, Dec and L,B of the same point in the sky

    ra, dec = (125.6, -75.3)
    l, b = (288.44190139183564, -20.717313145391525)

    # This should throw as we are using Powerlaw instead of Powerlaw()
    with pytest.raises(TypeError):

        _ = PointSource("my_source", ra, dec, Powerlaw)

    # Init with RA, Dec

    point_source1 = PointSource('my_source',ra, dec, Powerlaw())

    assert point_source1.position.get_ra() == ra
    assert point_source1.position.get_dec() == dec

    assert abs(point_source1.position.get_l() - l) < 1e-7
    assert abs(point_source1.position.get_b() - b) < 1e-7

    assert point_source1.position.ra.value == ra
    assert point_source1.position.dec.value == dec

    # Verify that the position is fixed by default
    assert point_source1.position.ra.fix
    assert point_source1.position.dec.fix

    # Init with l,b

    point_source2 = PointSource('my_source', l=l, b=b, spectral_shape=Powerlaw())

    assert point_source2.position.get_l() == l
    assert point_source2.position.get_b() == b

    assert abs(point_source2.position.get_ra() - ra) < 1e-7
    assert abs(point_source2.position.get_dec() - dec) < 1e-7

    assert point_source2.position.l.value == l
    assert point_source2.position.b.value == b

    # Verify that the position is fixed by default
    assert point_source2.position.l.fix
    assert point_source2.position.b.fix

    # Multi-component

    po1 = Powerlaw()
    po2 = Powerlaw()

    c1 = SpectralComponent("component1", po1)
    c2 = SpectralComponent("component2", po2)

    point_source3 = PointSource("test_source", ra, dec, components=[c1, c2])

    assert np.all(point_source3.spectrum.component1([1,2,3]) == po1([1,2,3]))
    assert np.all(point_source3.spectrum.component2([1,2,3]) == po2([1,2,3]))

    with pytest.raises(AssertionError):

        # Illegal RA

        _ = PointSource("test",720.0, -15.0, components=[c1,c2])

    with pytest.raises(AssertionError):
        # Illegal Dec

        _ = PointSource("test", 120.0, 180.0, components=[c1, c2])

    with pytest.raises(AssertionError):
        # Illegal l

        _ = PointSource("test", l=-195, b=-15.0, components=[c1, c2])

    with pytest.raises(AssertionError):
        # Illegal b

        _ = PointSource("test", l=120.0, b=-180.0, components=[c1, c2])


def test_call():

    # Multi-component

    po1 = Powerlaw()
    po2 = Powerlaw()

    c1 = SpectralComponent("component1", po1)
    c2 = SpectralComponent("component2", po2)

    point_source = PointSource("test_source", 125.4, -22.3, components=[c1, c2])

    assert np.all(point_source.spectrum.component1([1, 2, 3]) == po1([1, 2, 3]))
    assert np.all(point_source.spectrum.component2([1, 2, 3]) == po2([1, 2, 3]))

    one = point_source.spectrum.component1([1, 2, 3])
    two = point_source.spectrum.component2([1, 2, 3])

    assert np.all( np.abs(one + two - point_source([1,2,3])) == 0 )


def test_call_with_units():

    po = Powerlaw()

    result = po(1.0)

    assert result.ndim == 0

    with pytest.raises(AssertionError):

        # This raises because the units of the function have not been set up

        _ = po(1.0 * u.keV)

    # Use the function as a spectrum
    ps = PointSource("test",0,0,po)

    result = po(1.0 * u.keV)

    assert isinstance(result, u.Quantity)

    result = po(np.array([1,2,3])* u.keV)

    assert isinstance(result, u.Quantity)

    # Now test all the functions
    def test_one(class_type):

        instance = class_type()
        
        if not instance.is_prior:

            # if we have fixed x_units then we will use those
            # in the test

            if instance.has_fixed_units():

                x_unit_to_use = instance.fixed_units[0]

            else:

                x_unit_to_use = u.keV
                
            # Use the function as a spectrum
            ps = PointSource("test", 0, 0, instance)

            if instance.name in [ "Synchrotron", "_ComplexTestFunction" ]:
                particleSource = ParticleSource("particles", Powerlaw())
                instance.set_particle_distribution(particleSource.spectrum.main.shape)


            # elif instance.name in ["PhAbs", "TbAbs"]:

            #     instance
                

            result = ps(1.0)

            assert isinstance(result, float)

            result = ps(1.0 * x_unit_to_use)

            assert isinstance(result, u.Quantity)

            result = ps(np.array([1, 2, 3]) * x_unit_to_use)

            assert isinstance(result, u.Quantity)
            
            if instance.name in [ "Synchrotron", "_ComplexTestFunction" ]:
              model = Model( particleSource, ps)
            else:
              model = Model( ps )
            
            new_model = clone_model( model )
            
            new_result =  new_model["test"](np.array([1, 2, 3]) * x_unit_to_use)
            
            assert np.all(new_result==result)

            model.save("__test.yml", overwrite=True)
            
            new_model = load_model("__test.yml")

            new_result =  new_model["test"](np.array([1, 2, 3]) * x_unit_to_use)
            
            assert np.all(new_result==result)

        else:

            print ('Skipping prior function')



    for key in _known_functions:

        this_function = _known_functions[key]

        # Test only the power law of XSpec, which is the only one we know we can test at 1 keV

        if key.find("XS")==0 and key != "XS_powerlaw" or (key in _multiplicative_models):

            # An XSpec model. Test it only if it's a power law (the others might need other parameters during
            # initialization)

            continue

        if key.find("TemplateModel")==0:

            # The TemplateModel function has its own test

            continue

#        if key.find("Synchrotron")==0:

            # Naima Synchtron function should have its own test

#            continue

        if this_function._n_dim == 1:

            print("testing %s ..." % key)

            test_one(_known_functions[key])


def test_call_with_composite_function_with_units():

    def one_test(spectrum):

        print("Testing %s" % spectrum.expression)

        # # if we have fixed x_units then we will use those
        # # in the test
        #
        # if spectrum.expression.has_fixed_units():
        #
        #     x_unit_to_use, y_unit_to_use = spectrum.expression.fixed_units[0]
        #
        # else:

        x_unit_to_use = u.keV

        pts = PointSource("test", ra=0, dec=0, spectral_shape=spectrum)

        res = pts([100, 200] * x_unit_to_use)

        # This will fail if the units are wrong
        res.to(old_div(1, (u.keV * u.cm**2 * u.s)))

    # Test a simple composition

    spectrum = Powerlaw() * Exponential_cutoff()

    one_test(spectrum)

    spectrum = Band() + Blackbody()

    one_test(spectrum)

    # Test a more complicate composition

    spectrum = Powerlaw() * Exponential_cutoff() + Blackbody()

    one_test(spectrum)

    spectrum = Powerlaw() * Exponential_cutoff() * Exponential_cutoff() + Blackbody()

    one_test(spectrum)

    # test the absorption models

    if has_abs_models:
    
        spectrum = PhAbs() * Powerlaw()


        one_test(spectrum)

        spectrum = TbAbs() * Powerlaw()

        one_test(spectrum)


        spectrum = WAbs() * Powerlaw()

        one_test(spectrum)

    
    if has_xspec:

        spectrum = XS_phabs() * Powerlaw()

        one_test(spectrum)

        spectrum = XS_phabs() * XS_powerlaw()

        one_test(spectrum)

        spectrum = XS_phabs() * XS_powerlaw() * XS_phabs()

        one_test(spectrum)

        spectrum = XS_phabs() * XS_powerlaw() * XS_phabs() + Blackbody()

        one_test(spectrum)

        spectrum = XS_phabs() * XS_powerlaw() * XS_phabs() + XS_powerlaw()

        one_test(spectrum)
        
    if has_ebl:
    
        spectrum = Powerlaw() * EBLattenuation()
        
        one_test(spectrum)


def test_free_param():

    spectrum = Log_parabola()
    source = PointSource("test_source", ra=123.4, dec=56.7, spectral_shape=spectrum)

    parameters = [spectrum.alpha, spectrum.beta, spectrum.piv, spectrum.K, source.position.ra, source.position.dec]

    for param in parameters:
        param.free = False

    assert len(source.free_parameters) == 0

    for i, param in enumerate(parameters):
        param.free = True
        assert len(source.free_parameters) == i+1




    def test_local_deriv():

        p = Powerlaw(index=-2.)


        npt.assert_allclose(-2., p.local_spectral_index(np.logspace(1,3,10)))
        

        



from __future__ import division

import hashlib

import astropy.units as u
import numpy as np
from astropy.coordinates import ICRS, BaseCoordinateFrame, SkyCoord
from astropy.io import fits
from past.utils import old_div
from scipy.interpolate import RegularGridInterpolator

from astromodels.functions.function import Function3D, FunctionMeta
from astromodels.utils.angular_distance import angular_distance_fast


class Continuous_injection_diffusion_ellipse(Function3D, metaclass=FunctionMeta):
    r"""
        description :

            Positron and electrons diffusing away from the accelerator

        latex : $\left(\frac{180^\circ}{\pi}\right)^2 \frac{1.2154}{\sqrt{\pi^3} r_{\rm diff} ({\rm angsep} ({\rm x, y, lon_0, lat_0})+0.06 r_{\rm diff} )} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm x, y, lon_0, lat_0})}{r_{\rm diff} ^2} \right)$

        parameters :

            lon0 :

                desc : Longitude of the center of the source
                initial value : 0.0
                min : 0.0
                max : 360.0

            lat0 :

                desc : Latitude of the center of the source
                initial value : 0.0
                min : -90.0
                max : 90.0

            rdiff0 :

                desc : Projected diffusion radius. The maximum allowed value is used to define the truncation radius.
                initial value : 1.0
                min : 0
                max : 20

            delta :

                desc : index for the diffusion coefficient
                initial value : 0.5
                min : 0.3
                max : 0.6
                fix : yes

            b :

                desc : b field strength in uG
                initial value : 3
                min : 1
                max : 10.
                fix : yes

            piv :

                desc : Pivot for the diffusion radius
                initial value : 2e10
                min : 0
                fix : yes

            piv2 :

                desc : Pivot for converting gamma energy to electron energy (always be 1 TeV)
                initial value : 1e9
                min : 0
                fix : yes

            incl :

                desc : inclination of semimajoraxis to a line of constant latitude
                initial value : 0.0
                min : -90.0
                max : 90.0
                fix : yes

            elongation :

                desc : elongation of the ellipse (b/a)
                initial value : 1.
                min : 0.1
                max : 10.

        """

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.rdiff0.unit = x_unit

        # Delta is of course unitless

        self.delta.unit = u.dimensionless_unscaled
        self.b.unit = u.dimensionless_unscaled
        self.incl.unit = x_unit
        self.elongation.unit = u.dimensionless_unscaled

        # Piv has the same unit as energy (which is z)

        self.piv.unit = z_unit
        self.piv2.unit = z_unit

    def evaluate(self, x, y, z, lon0, lat0, rdiff0, delta, b, piv, piv2, incl, elongation):

        lon, lat = x, y
        energy = z

        # energy in kev -> TeV.
        # NOTE: the use of piv2 is necessary to preserve dimensional correctness: the logarithm can only be taken
        # of a dimensionless quantity, so there must be a pivot there.

        e_energy_piv2 = 17. * \
            np.power(old_div(energy, piv2), 0.54 + 0.046 *
                     np.log10(old_div(energy, piv2)))
        e_piv_piv2 = 17. * \
            np.power(old_div(piv, piv2), 0.54 + 0.046 *
                     np.log10(old_div(piv, piv2)))

        try:

            rdiff_a = rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), old_div((delta - 1.), 2.)) * \
                np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 * np.power(1. + 0.0107 * e_piv_piv2, -1.5)) / \
                np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 *
                        np.power(1. + 0.0107 * e_energy_piv2, -1.5))

        except ValueError:

            # This happens when using units, because astropy.units fails with the message:
            # "ValueError: Quantities and Units may only be raised to a scalar power"

            # Work around the problem with this loop, which is slow but using units is only for testing purposes or
            # single calls, so it shouldn't matter too much
            rdiff_a = np.array([(rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), x)).value for x in (delta - 1.) / 2. * np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 * np.power(1. + 0.0107 * e_piv_piv2, -1.5)) /
                                np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 * np.power(1. + 0.0107 * e_energy_piv2, -1.5))]) * rdiff0.unit

        rdiff_b = rdiff_a * elongation

        pi = np.pi

        angsep = angular_distance_fast(lon, lat, lon0, lat0)
        ang = np.arctan2(lat - lat0, (lon - lon0) *
                         np.cos(lat0 * np.pi / 180.))

        theta = np.arctan2(old_div(np.sin(ang-incl*np.pi/180.),
                                   elongation), np.cos(ang-incl*np.pi/180.))

        rdiffs_a, thetas = np.meshgrid(rdiff_a, theta)
        rdiffs_b, angseps = np.meshgrid(rdiff_b, angsep)

        rdiffs = np.sqrt(rdiffs_a ** 2 * np.cos(thetas) **
                         2 + rdiffs_b ** 2 * np.sin(thetas) ** 2)

        results = np.power(old_div(180.0, pi), 2) * 1.22 / (pi * np.sqrt(pi) * rdiffs_a * np.sqrt(
            elongation) * (angseps + 0.06 * rdiffs)) * np.exp(old_div(-np.power(angseps, 2), rdiffs ** 2))

        return results

    def get_boundaries(self):

        # Truncate the function at the max of rdiff allowed

        maximum_rdiff = self.rdiff0.max_value

        min_latitude = max(-90., self.lat0.value - maximum_rdiff)
        max_latitude = min(90., self.lat0.value + maximum_rdiff)

        max_abs_lat = max(np.absolute(min_latitude), np.absolute(max_latitude))

        if max_abs_lat > 89. or old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.)) >= 180.:

            min_longitude = 0.
            max_longitude = 360.

        else:

            min_longitude = self.lon0.value - \
                old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))
            max_longitude = self.lon0.value + \
                old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))

            if min_longitude < 0.:

                min_longitude += 360.

            elif max_longitude > 360.:

                max_longitude -= 360.

        return (min_longitude, max_longitude), (min_latitude, max_latitude)

    def get_total_spatial_integral(self, z=None):
        """
        Returns the total integral (for 2D functions) or the integral over the spatial components (for 3D functions).
        needs to be implemented in subclasses.

        :return: an array of values of the integral (same dimension as z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)


class Continuous_injection_diffusion(Function3D, metaclass=FunctionMeta):
    r"""
        description :

            Positron and electrons diffusing away from the accelerator

        latex : $\left(\frac{180^\circ}{\pi}\right)^2 \frac{1.2154}{\sqrt{\pi^3} r_{\rm diff} ({\rm angsep} ({\rm x, y, lon_0, lat_0})+0.06 r_{\rm diff} )} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm x, y, lon_0, lat_0})}{r_{\rm diff} ^2} \right)$

        parameters :

            lon0 :

                desc : Longitude of the center of the source
                initial value : 0.0
                min : 0.0
                max : 360.0

            lat0 :

                desc : Latitude of the center of the source
                initial value : 0.0
                min : -90.0
                max : 90.0

            rdiff0 :

                desc : Projected diffusion radius limited by the cooling time. The maximum allowed value is used to define the truncation radius.
                initial value : 1.0
                min : 0
                max : 20

            rinj :

                desc : Ratio of diffusion radius limited by the injection time over rdiff0. The maximum allowed value is used to define the truncation radius.
                initial value : 100.0
                min : 0
                max : 200
                fix : yes

            delta :

                desc : index for the diffusion coefficient
                initial value : 0.5
                min : 0.3
                max : 0.6
                fix : yes

            b :

                desc : b field strength in uG
                initial value : 3
                min : 1
                max : 10.
                fix : yes

            piv :

                desc : Pivot for the diffusion radius
                initial value : 2e10
                min : 0
                fix : yes

            piv2 :
                desc : Pivot for converting gamma energy to electron energy (always be 1 TeV)
                initial value : 1e9
                min : 0
                fix : yes

        """

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.rdiff0.unit = x_unit
        self.rinj.unit = u.dimensionless_unscaled

        # Delta is of course unitless

        self.delta.unit = u.dimensionless_unscaled
        self.b.unit = u.dimensionless_unscaled

        # Piv has the same unit as energy (which is z)

        self.piv.unit = z_unit
        self.piv2.unit = z_unit

    def evaluate(self, x, y, z, lon0, lat0, rdiff0, rinj, delta, b, piv, piv2):

        lon, lat = x, y
        energy = z

        # energy in kev -> TeV.
        # NOTE: the use of piv2 is necessary to preserve dimensional correctness: the logarithm can only be taken
        # of a dimensionless quantity, so there must be a pivot there.

        e_energy_piv2 = 17. * \
            np.power(old_div(energy, piv2), 0.54 + 0.046 *
                     np.log10(old_div(energy, piv2)))
        e_piv_piv2 = 17. * \
            np.power(old_div(piv, piv2), 0.54 + 0.046 *
                     np.log10(old_div(piv, piv2)))

        rdiff_c = rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), old_div((delta - 1.), 2.)) * \
            np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 * np.power(1. + 0.0107 * e_piv_piv2, -1.5)) / \
            np.sqrt(b * b / 8. / np.pi * 0.624 + 0.26 *
                    np.power(1. + 0.0107 * e_energy_piv2, -1.5))

        rdiff_i = rdiff0 * rinj * \
            np.power(old_div(e_energy_piv2, e_piv_piv2), old_div(delta, 2.))

        rdiff = np.minimum(rdiff_c, rdiff_i)

        angsep = angular_distance_fast(lon, lat, lon0, lat0)

        pi = np.pi

        rdiffs, angseps = np.meshgrid(rdiff, angsep)

        return np.power(old_div(180.0, pi), 2) * 1.2154 / (pi * np.sqrt(pi) * rdiffs * (angseps + 0.06 * rdiffs)) * \
            np.exp(old_div(-np.power(angseps, 2), rdiffs ** 2))

    def get_boundaries(self):

        # Truncate the function at the max of rdiff allowed

        maximum_rdiff = self.rdiff0.max_value

        min_latitude = max(-90., self.lat0.value - maximum_rdiff)
        max_latitude = min(90., self.lat0.value + maximum_rdiff)

        max_abs_lat = max(np.absolute(min_latitude), np.absolute(max_latitude))

        if max_abs_lat > 89. or old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.)) >= 180.:

            min_longitude = 0.
            max_longitude = 360.

        else:

            min_longitude = self.lon0.value - \
                old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))
            max_longitude = self.lon0.value + \
                old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))

            if min_longitude < 0.:

                min_longitude += 360.

            elif max_longitude > 360.:

                max_longitude -= 360.

        return (min_longitude, max_longitude), (min_latitude, max_latitude)

    def get_total_spatial_integral(self, z=None):
        """
        Returns the total integral (for 2D functions) or the integral over the spatial components (for 3D functions).
        needs to be implemented in subclasses.

        :return: an array of values of the integral (same dimension as z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)


class Continuous_injection_diffusion_legacy(Function3D, metaclass=FunctionMeta):
    r"""
        description :

            Positron and electrons diffusing away from the accelerator

        latex : $\left(\frac{180^\circ}{\pi}\right)^2 \frac{1.2154}{\sqrt{\pi^3} r_{\rm diff} ({\rm angsep} ({\rm x, y, lon_0, lat_0})+0.06 r_{\rm diff} )} \, {\rm exp}\left(-\frac{{\rm angsep}^2 ({\rm x, y, lon_0, lat_0})}{r_{\rm diff} ^2} \right)$

        parameters :

            lon0 :

                desc : Longitude of the center of the source
                initial value : 0.0
                min : 0.0
                max : 360.0

            lat0 :

                desc : Latitude of the center of the source
                initial value : 0.0
                min : -90.0
                max : 90.0

            rdiff0 :

                desc : Projected diffusion radius. The maximum allowed value is used to define the truncation radius.
                initial value : 1.0
                min : 0
                max : 20

            delta :

                desc : index for the diffusion coefficient
                initial value : 0.5
                min : 0.3
                max : 0.6
                fix : yes

            uratio :

                desc : ratio between u_cmb and u_B
                initial value : 0.5
                min : 0.01
                max : 100.
                fix : yes

            piv :

                desc : Pivot for the diffusion radius
                initial value : 2e10
                min : 0
                fix : yes

            piv2 :
                desc : Pivot for converting gamma energy to electron energy (always be 1 TeV)
                initial value : 1e9
                min : 0
                fix : yes

        """

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        # lon0 and lat0 and rdiff have most probably all units of degrees. However,
        # let's set them up here just to save for the possibility of using the
        # formula with other units (although it is probably never going to happen)

        self.lon0.unit = x_unit
        self.lat0.unit = y_unit
        self.rdiff0.unit = x_unit

        # Delta is of course unitless

        self.delta.unit = u.dimensionless_unscaled
        self.uratio.unit = u.dimensionless_unscaled

        # Piv has the same unit as energy (which is z)

        self.piv.unit = z_unit
        self.piv2.unit = z_unit

    def evaluate(self, x, y, z, lon0, lat0, rdiff0, delta, uratio, piv, piv2):

        lon, lat = x, y
        energy = z

        # energy in kev -> TeV.
        # NOTE: the use of piv2 is necessary to preserve dimensional correctness: the logarithm can only be taken
        # of a dimensionless quantity, so there must be a pivot there.

        e_energy_piv2 = 17. * \
            np.power(old_div(energy, piv2), 0.54 + 0.046 *
                     np.log10(old_div(energy, piv2)))
        e_piv_piv2 = 17. * \
            np.power(old_div(piv, piv2), 0.54 + 0.046 *
                     np.log10(old_div(piv, piv2)))

        try:

            rdiff = rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), old_div((delta - 1.), 2.)) * \
                np.sqrt(1. + uratio * np.power(1. + 0.0107 * e_piv_piv2, -1.5)) / \
                np.sqrt(1. + uratio * np.power(1. +
                                               0.0107 * e_energy_piv2, -1.5))

        except ValueError:

            # This happens when using units, because astropy.units fails with the message:
            # "ValueError: Quantities and Units may only be raised to a scalar power"

            # Work around the problem with this loop, which is slow but using units is only for testing purposes or
            # single calls, so it shouldn't matter too much
            rdiff = np.array([(rdiff0 * np.power(old_div(e_energy_piv2, e_piv_piv2), x)).value for x in (delta - 1.) / 2. * np.sqrt(1. + uratio * np.power(1. + 0.0107 * e_piv_piv2, -1.5)) /
                              np.sqrt(1. + uratio * np.power(1. + 0.0107 * e_energy_piv2, -1.5))]) * rdiff0.unit

        angsep = angular_distance_fast(lon, lat, lon0, lat0)

        pi = np.pi

        rdiffs, angseps = np.meshgrid(rdiff, angsep)

        return np.power(old_div(180.0, pi), 2) * 1.2154 / (pi * np.sqrt(pi) * rdiffs * (angseps + 0.06 * rdiffs)) * \
            np.exp(old_div(-np.power(angseps, 2), rdiffs ** 2))

    def get_boundaries(self):

        # Truncate the function at the max of rdiff allowed

        maximum_rdiff = self.rdiff0.max_value

        min_latitude = max(-90., self.lat0.value - maximum_rdiff)
        max_latitude = min(90., self.lat0.value + maximum_rdiff)

        max_abs_lat = max(np.absolute(min_latitude), np.absolute(max_latitude))

        if max_abs_lat > 89. or old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.)) >= 180.:

            min_longitude = 0.
            max_longitude = 360.

        else:

            min_longitude = self.lon0.value - \
                old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))
            max_longitude = self.lon0.value + \
                old_div(maximum_rdiff, np.cos(max_abs_lat * np.pi / 180.))

            if min_longitude < 0.:

                min_longitude += 360.

            elif max_longitude > 360.:

                max_longitude -= 360.

        return (min_longitude, max_longitude), (min_latitude, max_latitude)

    def get_total_spatial_integral(self, z=None):
        """
        Returns the total integral (for 2D functions) or the integral over the spatial components (for 3D functions).
        needs to be implemented in subclasses.

        :return: an array of values of the integral (same dimension as z).
        """

        if isinstance(z, u.Quantity):
            z = z.value
        return np.ones_like(z)


class GalPropTemplate_3D(Function3D):
    r"""
        description :

            Use a 3D template that has morphology and flux information.
            GalProp, DRAGON or a similar model in fits format would work. 
            Only parameter is a normalization factor. 

        latex : $ K $

        parameters :

            K :

                desc : normalization
                initial value : 1
                fix : yes

            hash :

                desc : hash of model map [needed for memoization]
                initial value : 1
                fix : yes         

    """

    __metaclass__ = FunctionMeta

    def _set_units(self, x_unit, y_unit, z_unit, w_unit):

        self.K.unit = (u.MeV * u.cm**2 * u.s * u.sr) ** (-1)

    def _setup(self):

        self._frame = ICRS()
        self._map = None
        self._fitsfile = None
        self._interpmap = None

    def set_frame(self, new_frame):
        """
        Set a new frame for the coordinates (the default is ICRS J2000)

        :param new_frame: a coordinate frame from astropy
        :return: (none)
        """
        assert isinstance(new_frame, BaseCoordinateFrame)

        self._frame = new_frame

    def load_file(self, fitsfile, phi1, phi2, theta1, theta2, galactic=False, ihdu=0):

        if fitsfile is None:
            raise RuntimeError(
                "Need to specify a fits file with a template map.")

        self._fitsfile = fitsfile
        p1, p2, t1, t2 = self.define_region(
            phi1, phi2, theta1, theta2, galactic)
        self.ramin = p1
        self.ramax = p2
        self.decmin = t1
        self.decmax = t2

        with fits.open(self._fitsfile) as f:

            self._delLon = f[ihdu].header['CDELT1']
            self._delLat = f[ihdu].header['CDELT2']
            self._delEn = f[ihdu].header['CDELT3']
            self._refLon = f[ihdu].header['CRVAL1']
            self._refLat = f[ihdu].header['CRVAL2']
            self._refEn = f[ihdu].header['CRVAL3']  # values in log10
            self._map = f[ihdu].data
            self._nl = f[ihdu].header['NAXIS1']  # longitude
            self._nb = f[ihdu].header['NAXIS2']  # latitude
            self._ne = f[ihdu].header['NAXIS3']  # energy

            # Create the function for the interpolation
            self._L = np.linspace(
                self._refLon, self._refLon+(self._nl-1)*self._delLon, self._nl)
            self._B = np.linspace(
                self._refLat, self._refLat+(self._nb-1)*self._delLat, self._nb)
            self._E = np.linspace(
                self._refEn, self._refEn+(self._ne-1)*self._delEn, self._ne)
            for i in range(len(self._E)):
                # Map units in Mev / cm^2 s sr, changing to 1 / MeV cm^2 s sr
                self._map[i] = self._map[i] / \
                    (np.power(10, self._E[i])*np.power(10, self._E[i]))
                self._map[i] = (np.fliplr(self._map[i]))
            self._F = RegularGridInterpolator(
                (self._E, self._B, self._L), self._map, bounds_error=False)

            h = hashlib.sha224()
            h.update(self._map)
            self.hash = int(h.hexdigest(), 16)

    def to_dict(self, minimal=False):

        data = super(Function3D, self).to_dict(minimal)

        if not minimal:

            data['extra_setup'] = {
                "_fitsfile": self._fitsfile,
                "_frame": self._frame,
                "ramin": self.ramin,
                "ramax": self.ramax,
                "decmin": self.decmin,
                "decmax": self.decmax
            }

        return data

    def which_model_file(self):
        return self._fitsfile

    def evaluate(self, x, y, z, K, hash):

        if self._map is None:

            self.load_file(self._fitsfile, self.ramin, self.ramax,
                           self.decmin, self.decmax, False, ihdu=0)

        # Interpolated values can be cached since we are fitting the constant K
        if self._interpmap is None:

            # We assume x and y are R.A. and Dec
            _coord = SkyCoord(ra=x, dec=y, frame=self._frame, unit="deg")
            b = _coord.transform_to('galactic').b.value
            l = _coord.transform_to('galactic').l.value
            lon = l
            lat = b

            # transform energy from keV to MeV. Galprop Model starts at 100 MeV
            energy = np.log10(z)-np.log10((u.MeV.to('keV')/u.keV).value)

            if lon.size != lat.size:
                raise AttributeError("Lon and Lat should be the same size")
            f = np.zeros([lon.size, energy.size])
            E0 = self._refEn
            Ef = self._refEn + (self._ne-1)*self._delEn

            # fix longitude
            shift = np.where(lon > 180.)
            lon[shift] = 180 - lon[shift]

            for i in range(energy.size):
                e = np.repeat(energy[i], len(lon))
                try:
                    f[:, i] = self._F(zip(e, lat, lon))
                except ValueError:
                    pass
            bad_idx = np.isnan(f)
            f[bad_idx] = 0
            bad_idx = np.isinf(f)
            f[bad_idx] = 0
            assert np.all(np.isfinite(f)), "some interpolated values are wrong"
            self._interpmap = f

        # (1000 is to change from MeV to KeV)
        A = np.multiply(K, self._interpmap/1000.)
        return A

    def define_region(self, a, b, c, d, galactic=False):
        if galactic:
            lmin = a
            lmax = b
            bmin = c
            bmax = d

            _coord = SkyCoord(l=[lmin, lmin, lmax, lmax], b=[
                              bmin, bmax, bmax, bmin], frame='galactic', unit='deg')

            ramin = min(_coord.transform_to('icrs').ra.value)
            ramax = max(_coord.transform_to('icrs').ra.value)
            decmin = min(_coord.transform_to('icrs').dec.value)
            decmax = max(_coord.transform_to('icrs').dec.value)

        else:
            ramin = a
            ramax = b
            decmin = c
            decmax = d
        return ramin, ramax, decmin, decmax

    def get_boundaries(self):
        min_longitude = self.ramin
        max_longitude = self.ramax
        min_latitude = self.decmin
        max_latitude = self.decmax
        return (min_longitude, max_longitude), (min_latitude, max_latitude)



__author__ = 'giacomov'

from .point_source import PointSource
from .extended_source import ExtendedSource
from .particle_source import ParticleSource
from .source import Source, SourceType

__all__ = ["PointSource", "ExtendedSource", "ParticleSource"]





# If threeml is not installed, we create our own log,
# otherwise, just print to the 3ML one!


import logging
import logging.handlers as handlers
import sys
from hashlib import sha256
from pathlib import Path

from astromodels.utils.configuration import astromodels_config

from .sys_tools import colored, supports_color

try:

    from threeML.config.config import threeML_config

    has_threeml = True

except:

    has_threeml = False


def get_path_of_log_dir():

    # we use the 3ML log path to simplify things
    # a more clever solution could be found

    if has_threeml:

        user_log: Path = Path(threeML_config.logging.path).expanduser()

    else:

        user_log: Path = Path(astromodels_config.logging.path).expanduser()

    # Create it if doesn't exist
    if not user_log.exists():

        user_log.mkdir(parents=True)

    return user_log


_log_file_names = ["usr.log", "dev.log"]


def get_path_of_log_file(log_file: str) -> Path:
    """
    returns the path of the log files
    """
    assert log_file in _log_file_names, f"{log_file} is not one of {_log_file_names}"

    return get_path_of_log_dir() / log_file


DEFAULT_LOG_COLORS = {
    'DEBUG': 'blue',
    'INFO': 'green',
    'WARNING': 'purple',
    'ERROR': 'red',
    'CRITICAL': 'bold_red',

}


def esc(*x):
    """Create escaped code from format code"""
    return '\033[' + ';'.join(x) + 'm'


# The following coloured log logic is from
# https://github.com/borntyping/python-colorlog
# I dropped some features and removed the Python 2.7 compatibility

ESCAPE_CODES = {'reset': esc('0'), 'bold': esc('01'), 'thin': esc('02')}

COLORS = ['black', 'red', 'green', 'yellow', 'blue', 'purple', 'cyan', 'white']

PREFIXES = [
    # Foreground without prefix
    ('3', ''),
    ('01;3', 'bold_'),
    ('02;3', 'thin_'),

    # Foreground with fg_ prefix
    ('3', 'fg_'),
    ('01;3', 'fg_bold_'),
    ('02;3', 'fg_thin_'),

    # Background with bg_ prefix - bold/light works differently
    ('4', 'bg_'),
    ('10', 'bg_bold_'),
]

for _prefix, _prefix_name in PREFIXES:
    for _code, _name in enumerate(COLORS):
        ESCAPE_CODES[_prefix_name + _name] = esc(_prefix + str(_code))


def parse_colors(sequence):
    """Return escape codes from a color sequence."""
    return ''.join(ESCAPE_CODES[n] for n in sequence.split(',') if n)


class ColoredRecord:
    """
    Wraps a LogRecord, adding named escape codes to the internal dict.
    The internal dict is used when formatting the message (by the PercentStyle,
    StrFormatStyle, and StringTemplateStyle classes).
    """

    def __init__(self, record):
        """Add attributes from the escape_codes dict and the record."""
        self.__dict__.update(ESCAPE_CODES)
        self.__dict__.update(record.__dict__)
        self.__record = record

    def __getattr__(self, name):
        return getattr(self.__record, name)


class ColoredFormatter(logging.Formatter):
    """
    A formatter that allows colors to be placed in the format string.
    Intended to help in creating more readable logging output.
    Based on https://github.com/borntyping/python-colorlog
    """

    def __init__(self,
                 fmt,
                 datefmt=None,
                 style='%',
                 log_colors=None,
                 reset=True,
                 secondary_log_colors=None):
        """
        Set the format and colors the ColouredFormatter will use.
        The ``fmt``, ``datefmt`` and ``style`` args are passed on to the
        ``logging.Formatter`` constructor.
        The ``secondary_log_colors`` argument can be used to create additional
        ``log_color`` attributes. Each key in the dictionary will set
        ``{key}_log_color``, using the value to select from a different
        ``log_colors`` set.
        :Parameters:
        - fmt (str): The format string to use
        - datefmt (str): A format string for the date
        - log_colors (dict):
            A mapping of log level names to color names
        - reset (bool):
            Implicitly append a color reset to all records unless False
        - style ('%' or '{' or '$'):
            The format style to use. (*No meaning prior to Python 3.2.*)
        - secondary_log_colors (dict):
            Map secondary ``log_color`` attributes. (*New in version 2.6.*)
        """
        super(ColoredFormatter, self).__init__(fmt, datefmt, style)

        self.log_colors = (log_colors
                           if log_colors is not None else DEFAULT_LOG_COLORS)
        self.secondary_log_colors = secondary_log_colors
        self.reset = reset

    def format(self, record):
        """Format a message from a record object."""
        record = ColoredRecord(record)
        record.log_color = escape_codes(self.log_colors, record.levelname)

        if self.secondary_log_colors:
            for name, log_colors in self.secondary_log_colors.items():
                color = escape_codes(log_colors, record.levelname)
                setattr(record, name + '_log_color', color)

        message = super(ColoredFormatter, self).format(record)

        if self.reset and not message.endswith(ESCAPE_CODES['reset']):
            message += ESCAPE_CODES['reset']

        return message


def escape_codes(log_colors, level_name):
    """Return escape codes from a ``log_colors`` dict."""
    return parse_colors(log_colors.get(level_name, ""))


def hash_coloured(text):
    """Return a ANSI coloured text based on its hash"""
    ansi_code = int(sha256(text.encode('utf-8')).hexdigest(), 16) % 230
    return colored(text, ansi_code=ansi_code)


def hash_coloured_escapes(text):
    """Return the ANSI hash colour prefix and suffix for a given text"""
    ansi_code = int(sha256(text.encode('utf-8')).hexdigest(), 16) % 230
    prefix, suffix = colored('SPLIT', ansi_code=ansi_code).split('SPLIT')
    return prefix, suffix



class LogFilter(object):
    def __init__(self, level):
        self.__level = level

    def filter(self, logRecord):
        return logRecord.levelno != self.__level


# now create the developer handler that rotates every day and keeps
# 10 days worth of backup
astromodels_dev_log_handler = handlers.TimedRotatingFileHandler(
    get_path_of_log_file("dev.log"), when="D", interval=1, backupCount=10
)

# lots of info written out

_dev_formatter = logging.Formatter(
    "%(asctime)s | %(name)s | %(levelname)s| %(funcName)s | %(lineno)d | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

astromodels_dev_log_handler.setFormatter(_dev_formatter)
astromodels_dev_log_handler.setLevel(logging.DEBUG)
# now set up the usr log which will save the info

astromodels_usr_log_handler = handlers.TimedRotatingFileHandler(
    get_path_of_log_file("usr.log"), when="D", interval=1, backupCount=10
)

astromodels_usr_log_handler.setLevel(logging.INFO)

# lots of info written out
_usr_formatter = logging.Formatter(
    "%(asctime)s | %(levelname)s | %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)

astromodels_usr_log_handler.setFormatter(_usr_formatter)

# now set up the console logger
name = "test"
if supports_color():
    prefix_1, suffix = hash_coloured_escapes(name)
    prefix_2, _ = hash_coloured_escapes(name + 'salt')
else:
    prefix_1, prefix_2, suffix = ('', '', '')

date_str = ''

_console_formatter = ColoredFormatter('[%(log_color)s%(levelname)-8s%(reset)s]'
                                      '%(log_color)s %(message)s',
                                      datefmt="%H:%M:%S",

                                      )

astromodels_console_log_handler = logging.StreamHandler(sys.stdout)
astromodels_console_log_handler.setFormatter(_console_formatter)
astromodels_console_log_handler.setLevel("INFO")

warning_filter = LogFilter(logging.WARNING)


def silence_warnings():
    """
    supress warning messages in console and file usr logs
    """

    astromodels_usr_log_handler.addFilter(warning_filter)
    astromodels_console_log_handler.addFilter(warning_filter)


def activate_warnings():
    """
    supress warning messages in console and file usr logs
    """

    astromodels_usr_log_handler.removeFilter(warning_filter)
    astromodels_console_log_handler.removeFilter(warning_filter)


def update_logging_level(level):

    astromodels_console_log_handler.setLevel(level)


def setup_logger(name):

    # A logger with name name will be created
    # and then add it to the print stream
    log = logging.getLogger(name)

    # this must be set to allow debug messages through
    log.setLevel(logging.DEBUG)

    # add the handlers

    log.addHandler(astromodels_dev_log_handler)

    log.addHandler(astromodels_console_log_handler)

    log.addHandler(astromodels_usr_log_handler)

    # we do not want to duplicate teh messages in the parents
    log.propagate = False

    return log



from builtins import object, str

__author__ = "giacomov"

import re
import warnings

from astromodels.core import (model, parameter, polarization, sky_direction,
                              spectral_component)
from astromodels.core.my_yaml import my_yaml
from astromodels.functions import function
from astromodels.sources import extended_source, particle_source, point_source
from astromodels.sources.source import SourceType
from astromodels.utils.logging import setup_logger

log = setup_logger(__name__)


class ModelIOError(IOError):
    pass


class ModelYAMLError(my_yaml.YAMLError):
    pass


class ModelSyntaxError(RuntimeError):
    pass


def load_model(filename):
    """
    Load a model from a file.

    :param filename: the name of the file containing the model
    :return: an instance of a Model
    """

    parser = ModelParser(filename)

    return parser.get_model()


def clone_model(model_instance):
    """
    Returns a copy of the given model with all objects cloned. This is equivalent to saving the model to
    a file and reload it, but it doesn't require writing or reading to/from disk. The original model is not touched.

    :param model: model to be cloned
    :return: a cloned copy of the given model
    """

    data = model_instance.to_dict_with_types()

    parser = ModelParser(model_dict=data)

    return parser.get_model()


def model_unpickler(state):

    return ModelParser(model_dict=state).get_model()


class ModelParser(object):
    def __init__(self, model_file=None, model_dict=None):

        assert (model_file is not None) or (model_dict is not None), (
            "You have to provide either a model file or a" "model dictionary"
        )

        if model_file is not None:

            # Read model file and deserialize into a dictionary

            try:

                with open(model_file) as f:

                    self._model_dict = my_yaml.load(f, Loader=my_yaml.FullLoader)

            except IOError:

                raise ModelIOError(
                    "File %s cannot be read. Check path and permissions for current user."
                    % model_file
                )

            except my_yaml.YAMLError:

                raise ModelYAMLError(
                    "Could not parse file %s. Check your syntax." % model_file
                )

        else:

            self._model_dict = model_dict

        self._parse()

    def _parse(self):

        # Traverse the dictionary and create all the needed classes

        # The first level is the source level

        self._sources = []
        self._independent_variables = []
        self._external_parameters = []
        self._links = []
        self._external_parameter_links = []
        self._extra_setups = []

        for source_or_var_name, source_or_var_definition in list(
            self._model_dict.items()
        ):

            if source_or_var_name.find("(IndependentVariable)") > 0:

                var_name = source_or_var_name.split("(")[0].replace(" ", "")

                this_parser = IndependentVariableParser(
                    var_name, source_or_var_definition
                )

                res = this_parser.get_variable()

                assert isinstance(res, parameter.IndependentVariable)

                self._independent_variables.append(res)

            elif source_or_var_name.find("(Parameter)") > 0:

                var_name = source_or_var_name.split("(")[0].replace(" ", "")

                this_parser = ParameterParser(var_name, source_or_var_definition)

                res = this_parser.get_variable()

                assert isinstance(res, parameter.Parameter)

                self._external_parameters.append(res)

                self._links.extend(this_parser.links)
            #                self._external_parameter_links.extend(this_parser.links)

            else:

                this_parser = SourceParser(source_or_var_name, source_or_var_definition)

                res = this_parser.get_source()

                assert (
                    isinstance(res, point_source.PointSource)
                    or isinstance(res, extended_source.ExtendedSource)
                    or isinstance(res, particle_source.ParticleSource)
                )

                self._sources.append(res)

                self._links.extend(this_parser.links)

                self._extra_setups.extend(this_parser.extra_setups)

    def get_model(self):

        # Instance the model with all the parsed sources

        new_model = model.Model(*self._sources)

        # Now set up IndependentVariable instances (if any)

        for independent_variable in self._independent_variables:

            new_model.add_independent_variable(independent_variable)

        # Now set up external parameters (if any)
        for parameter in self._external_parameters:

            new_model.add_external_parameter(parameter)

        # Now set up the links

        for link in self._links:

            path = link["parameter_path"]
            variable = link["variable"]
            law = link["law"]

            new_model[path].add_auxiliary_variable(new_model[variable], law)

        # Finally the extra_setups (if any)

        for extra_setup in self._extra_setups:

            path = extra_setup["function_path"]

            for property, value in list(extra_setup["extra_setup"].items()):

                # First, check to see if the we have a valid path in the new model.
                # If we aren't given a path, interpret it as being given a value.
                if value in new_model:
                    new_model[path].__setattr__(property, new_model[value])
                else:
                    new_model[path].__setattr__(property, value)

        return new_model


class IndependentVariableParser(object):
    def __init__(self, name, definition):

        self._variable = parameter.IndependentVariable(name, **definition)

    def get_variable(self):

        return self._variable


class ParameterParser(object):
    def __init__(self, name, definition):

        self._links = []

        # NOTE: this is triggered only for parameters outside of functions

        if "prior" in definition:

            # Need the create a function for the prior first

            try:

                function_name = list(definition["prior"].keys())[0]
                parameters_definition = definition["prior"][function_name]

            except KeyError:  # pragma: no cover

                raise ModelSyntaxError("The prior for parameter %s is malformed" % name)

            # parse the function

            shape_parser = ShapeParser(name)

            prior_instance = shape_parser.parse(
                name, function_name, parameters_definition
            )

            # Substitute the definition with the instance, so that the following constructor will work
            definition["prior"] = prior_instance

        # Check if this is a linked parameter, i.e., if 'value' is something like f(source.spectrum.powerlaw.index)

        matches = re.findall("""f\((.+)\)""", str(definition["value"]))

        if matches:

            # This is an expression which marks a parameter
            # with a link to another parameter (or an IndependentVariable such as time)

            # Get the variable
            linked_variable = matches[0]

            # Now get the law

            if "law" not in definition:  # pragma: no cover
                raise ModelSyntaxError(
                    "The parameter %s in function %s "
                    " is linked to %s but lacks a 'law' attribute"
                    % (name, function_name, linked_variable)
                )

            link_function_name = list(definition["law"].keys())[0]

            # ok, now we parse the linked parameter

            function_parser = ShapeParser(name)

            link_function_instance = function_parser.parse(
                name, link_function_name, definition["law"][link_function_name]
            )

            self._links.append(
                {
                    "parameter_path": name,
                    "law": link_function_instance,
                    "variable": linked_variable,
                }
            )

            # get rid of the 'law' entry

            definition.pop("law", None)

            # this parameter's value will be replaced later.
            # for now we just need to get rid of the f(param) entry

            definition["value"] = 1.0

        self._variable = parameter.Parameter(name, **definition)

    def get_variable(self):

        return self._variable

    @property
    def links(self):

        return self._links


class SourceParser(object):
    def __init__(self, source_name, source_definition):

        # Get the type of the source

        try:

            # Point source or extended source?

            source_type = re.findall(
                "\((%s|%s|%s)\)" % (SourceType.POINT_SOURCE, SourceType.EXTENDED_SOURCE, SourceType.PARTICLE_SOURCE),
                source_name,
            )[-1]

            
            
        except IndexError:  # pragma: no cover

            log.error("Don't recognize type for source '%s'. "
                "Valid types are '%s', '%s' or '%s'."
                % (source_name, SourceType.POINT_SOURCE, SourceType.EXTENDED_SOURCE, SourceType.PARTICLE_SOURCE))
            
            raise ModelSyntaxError()

        else:

            # Strip the source_type from the name

            source_name = source_name.split()[0]

        self._source_name = source_name

        # This will store the links (if any)
        self._links = []

        # This will store extra_setups (if any), used sometimes. For example, the function which uses naima
        # to make a synchrotron spectrum uses this to save and set up the particle distribution
        self._extra_setups = []

        if source_type == SourceType.POINT_SOURCE.value:

            self._parsed_source = self._parse_point_source(source_definition)

        elif source_type == SourceType.EXTENDED_SOURCE.value:

            self._parsed_source = self._parse_extended_source(source_definition)

        elif source_type == SourceType.PARTICLE_SOURCE.value:

            self._parsed_source = self._parse_particle_source(source_definition)

    @property
    def extra_setups(self):

        return self._extra_setups

    @property
    def links(self):

        return self._links

    def get_source(self):

        return self._parsed_source

    def _parse_particle_source(self, particle_source_definition):

        # Parse the spectral information

        try:

            spectrum = particle_source_definition["spectrum"]

        except KeyError:  # pragma: no cover

            raise ModelSyntaxError(
                "Point source %s is missing the 'spectrum' attribute"
                % self._source_name
            )

        components = []

        for component_name, component_definition in list(
            particle_source_definition["spectrum"].items()
        ):

            this_component = self._parse_spectral_component(
                component_name, component_definition
            )

            components.append(this_component)

        this_particle_source = particle_source.ParticleSource(
            self._source_name, components=components
        )

        return this_particle_source

    def _parse_point_source(self, pts_source_definition):

        # Parse the positional information

        try:

            position_definition = pts_source_definition["position"]

        except KeyError:  # pragma: no cover

            raise ModelSyntaxError(
                "Point source %s is missing the 'position' attribute"
                % self._source_name
            )

        this_sky_direction = self._parse_sky_direction(position_definition)

        # Parse the spectral information

        try:

            spectrum = pts_source_definition["spectrum"]

        except KeyError:  # pragma: no cover

            raise ModelSyntaxError(
                "Point source %s is missing the 'spectrum' attribute"
                % self._source_name
            )

        components = []

        for component_name, component_definition in list(
            pts_source_definition["spectrum"].items()
        ):

            try:

                this_component = self._parse_spectral_component(
                    component_name, component_definition
                )

                components.append(this_component)

            except:

                raise

        try:

            this_point_source = point_source.PointSource(
                self._source_name,
                sky_position=this_sky_direction,
                components=components,
            )

        except:

            raise

        return this_point_source

    def _parse_sky_direction(self, sky_direction_definition):

        # Instance the SkyDirection class using the coordinates provided

        coordinates = {}

        if "ra" in sky_direction_definition and "dec" in sky_direction_definition:

            par_parser = ParameterParser("ra", sky_direction_definition["ra"])

            ra = par_parser.get_variable()

            if ra.bounds == (None, None):
              ra.bounds = (0, 360)

            par_parser = ParameterParser("dec", sky_direction_definition["dec"])

            dec = par_parser.get_variable()
            
            if dec.bounds == (None, None):
              dec.bounds = (-90, 90)

            coordinates["ra"] = ra
            coordinates["dec"] = dec

        elif "l" in sky_direction_definition and "b" in sky_direction_definition:

            par_parser = ParameterParser("l", sky_direction_definition["l"])

            l = par_parser.get_variable()
            
            if l.bounds == (None, None):
              l.bounds = (0, 360)

            par_parser = ParameterParser("b", sky_direction_definition["b"])

            b = par_parser.get_variable()
            
            if b.bounds == (None, None):
              b.bounds = (-90, 90)

            coordinates["l"] = l
            coordinates["b"] = b

        else:  # pragma: no cover

            raise ModelSyntaxError(
                "Position specification for source %s has an invalid coordinate pair. "
                " You need to specify either 'ra' and 'dec', or 'l' and 'b'."
                % self._source_name
            )

        # Check if there is a equinox specification

        if "equinox" in sky_direction_definition:
            coordinates["equinox"] = sky_direction_definition["equinox"]

        try:

            this_sky_direction = sky_direction.SkyDirection(**coordinates)

        except sky_direction.WrongCoordinatePair:  # pragma: no cover

            raise ModelSyntaxError(
                "Position specification for source %s has an invalid coordinate pair"
                % self._source_name
            )

        return this_sky_direction

    def _parse_polarization(self, polarization_definititon):

        polarization_params = {}

        if "degree" in polarization_definititon and "angle" in polarization_definititon:

            par_parser = ParameterParser("degree", polarization_definititon["degree"])

            degree = par_parser.get_variable()

            degree.bounds = (0, 100)

            par_parser = ParameterParser("angle", polarization_definititon["angle"])

            angle = par_parser.get_variable()

            angle.bounds = (0, 180)

            this_polarization = polarization.LinearPolarization(
                angle=angle, degree=degree
            )

        elif (
            "I" in polarization_definititon
            and "U" in polarization_definititon
            and "Q" in polarization_definititon
            and "V" in polarization_definititon
        ):

            par_parser = ParameterParser("I", polarization_definititon["I"])

            I = par_parser.get_variable()

            I.bounds = (0, 1)

            par_parser = ParameterParser("U", polarization_definititon["U"])

            U = par_parser.get_variable()

            U.bounds = (0, 1)

            par_parser = ParameterParser("Q", polarization_definititon["Q"])

            Q = par_parser.get_variable()

            Q.bounds = (0, 1)

            par_parser = ParameterParser("V", polarization_definititon["V"])

            V = par_parser.get_variable()

            V.bounds = (0, 1)

            this_polarization = polarization.StokesPolarization(I=I, Q=Q, U=U, V=V)

        else:

            # just make a default polarization

            this_polarization = polarization.Polarization()
            # raise ModelSyntaxError("Polarization specification for source %s has an invalid parameters. "
            #                        " You need to specify either 'angle' and 'degree', or 'I' ,'Q', 'U' and 'V'."
            #                        % self._source_name)

        return this_polarization

    def _parse_spectral_component(self, component_name, component_definition):

        # Parse the shape definition, which is the first to occur

        try:

            function_name = list(component_definition.keys())[0]
            parameters_definition = component_definition[function_name]

        except KeyError:  # pragma: no cover

            raise ModelSyntaxError(
                "The component %s of source %s is malformed"
                % (component_name, self._source_name)
            )

        # parse the function

        shape_parser = ShapeParser(self._source_name)

        shape = shape_parser.parse(
            component_name, function_name, parameters_definition, is_spatial=False
        )

        # Get the links and extra setups, if any

        self._links.extend(shape_parser.links)
        self._extra_setups.extend(shape_parser.extra_setups)

        if "polarization" in component_definition:

            # get the polarization

            polarization_definition = component_definition["polarization"]

            this_polarization = self._parse_polarization(polarization_definition)

        else:

            this_polarization = polarization.Polarization()

        this_spectral_component = spectral_component.SpectralComponent(
            component_name, shape, this_polarization
        )

        return this_spectral_component

    def _parse_extended_source(self, ext_source_definition):

        # The first item in the dictionary is the definition of the extended shape
        name_of_spatial_shape = list(ext_source_definition.keys())[0]

        spatial_shape_parser = ShapeParser(self._source_name)

        spatial_shape = spatial_shape_parser.parse(
            "n.a.",
            name_of_spatial_shape,
            list(ext_source_definition.values())[0],
            is_spatial=True,
        )

        # Get the links and extra setups, if any

        self._links.extend(spatial_shape_parser.links)
        self._extra_setups.extend(spatial_shape_parser.extra_setups)

        # Parse the spectral information

        try:

            spectrum = ext_source_definition["spectrum"]

        except KeyError:  # pragma: no cover

            raise ModelSyntaxError(
                "Ext. source %s is missing the 'spectrum' attribute" % self._source_name
            )

        components = []

        for component_name, component_definition in list(
            ext_source_definition["spectrum"].items()
        ):
            this_component = self._parse_spectral_component(
                component_name, component_definition
            )

            components.append(this_component)

        this_ext_source = extended_source.ExtendedSource(
            self._source_name, spatial_shape, components=components
        )

        return this_ext_source


class ShapeParser(object):
    def __init__(self, source_name):

        self._source_name = source_name
        self._links = []
        self._extra_setups = []

    @property
    def links(self):

        return self._links

    @property
    def extra_setups(self):

        return self._extra_setups

    def parse(
        self, component_name, function_name, parameters_definition, is_spatial=False
    ):

        return self._parse_shape_definition(
            component_name, function_name, parameters_definition, is_spatial
        )

    @staticmethod
    def _fix(value):
        # Remove new lines where it shouldn't be any
        # Sometimes YAML add new lines in the middle of definitions,
        # such as in units
        return value.replace("\n", " ")

    def _parse_shape_definition(
        self, component_name, function_name, parameters_definition, is_spatial=False
    ):

        # Get the function

        if "expression" in parameters_definition:

            # This is a composite function
            function_instance = function.get_function(
                function_name, parameters_definition["expression"]
            )

        else:

            try:

                function_instance = function.get_function(function_name)

            except function.UnknownFunction:  # pragma: no cover

                raise ModelSyntaxError(
                    "Function %s, specified as shape for %s of source %s, is not a "
                    "known function"
                    % (function_name, component_name, self._source_name)
                )

        # Loop over the parameters of the function instance, instead of the specification,
        # so we can understand if there are parameters missing from the specification

        for parameter_name, _ in function_instance.parameters.items():

            try:

                this_definition = parameters_definition[parameter_name]

            except KeyError:  # pragma: no cover

                raise ModelSyntaxError(
                    "Function %s, specified as shape for %s of source %s, lacks "
                    "the definition for parameter %s"
                    % (function_name, component_name, self._source_name, parameter_name)
                )

            # Update the parameter. Note that the order is important, because trying to set the value before the
            # minimum and maximum could result in a error.

            # All these specifications are optional. If they are not present, then the default value
            # already contained in the instance of the function will be used

            # Ignore for a second the RuntimeWarning that is printed if the default value in the function definition
            # is outside the bounds defined here

            with warnings.catch_warnings():

                warnings.simplefilter("ignore", RuntimeWarning)

                if "min_value" in this_definition:
                    function_instance.parameters[
                        parameter_name
                    ].min_value = this_definition["min_value"]

                if "max_value" in this_definition:
                    function_instance.parameters[
                        parameter_name
                    ].max_value = this_definition["max_value"]

            if "delta" in this_definition:
                function_instance.parameters[parameter_name].delta = this_definition[
                    "delta"
                ]

            if "free" in this_definition:
                function_instance.parameters[parameter_name].free = this_definition[
                    "free"
                ]

            if "unit" in this_definition:
                function_instance.parameters[parameter_name].unit = self._fix(
                    this_definition["unit"]
                )

            # Now set the value, which must be present

            if "value" not in this_definition:  # pragma: no cover
                raise ModelSyntaxError(
                    "The parameter %s in function %s, specified as shape for %s "
                    "of source %s, lacks a 'value' attribute"
                    % (parameter_name, function_name, component_name, self._source_name)
                )

            # Check if this is a linked parameter, i.e., if 'value' is something like f(source.spectrum.powerlaw.index)

            matches = re.findall("""f\((.+)\)""", str(this_definition["value"]))

            if matches:

                # This is an expression which marks a parameter
                # with a link to another parameter (or an IndependentVariable such as time)

                # Get the variable
                linked_variable = matches[0]

                # Now get the law

                if "law" not in this_definition:  # pragma: no cover
                    raise ModelSyntaxError(
                        "The parameter %s in function %s, specified as shape for %s "
                        "of source %s, is linked to %s but lacks a 'law' attribute"
                        % (
                            parameter_name,
                            function_name,
                            component_name,
                            self._source_name,
                            linked_variable,
                        )
                    )

                link_function_name = list(this_definition["law"].keys())[0]

                link_function_instance = self._parse_shape_definition(
                    component_name,
                    link_function_name,
                    this_definition["law"][link_function_name],
                )

                if is_spatial:
                    path = ".".join([self._source_name, function_name, parameter_name])
                else:
                    path = ".".join(
                        [
                            self._source_name,
                            "spectrum",
                            component_name,
                            function_name,
                            parameter_name,
                        ]
                    )

                self._links.append(
                    {
                        "parameter_path": path,
                        "law": link_function_instance,
                        "variable": linked_variable,
                    }
                )

            else:

                # This is a normal (not linked) parameter

                function_instance.parameters[parameter_name].value = this_definition[
                    "value"
                ]

            # Setup the prior for this parameter, if it exists
            if "prior" in this_definition:
                # Get the function for this prior

                # A name to display in case of errors

                name_for_errors = (
                    "prior for %s" % function_instance.parameters[parameter_name].path
                )

                prior_function_name = list(this_definition["prior"].keys())[0]

                prior_function_definition = this_definition["prior"][
                    prior_function_name
                ]

                prior_function = self._parse_shape_definition(
                    name_for_errors, prior_function_name, prior_function_definition
                )

                # Set it as prior for current parameter

                function_instance.parameters[parameter_name].prior = prior_function

        # Now handle extra_setup if any
        if "extra_setup" in parameters_definition:

            if is_spatial:
                path = ".".join([self._source_name, function_name])
            else:
                path = ".".join(
                    [self._source_name, "spectrum", component_name, function_name]
                )

            self._extra_setups.append(
                {
                    "function_path": path,
                    "extra_setup": parameters_definition["extra_setup"],
                }
            )

        return function_instance
